/*
 *  pcixcom.c --- program to interactively work with the OMS PCIx card via
 *     the Linux driver (oms-pcix.ko).
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 *
 *  This version uses the ncurses library, in a fairly rudimentary way,
 *  to split the screen into an input area, where the user types, and
 *  an output area, where information from the PCIx controller card is
 *  displayed.  Output includes any text responses as well as notifications
 *  that flags have been set.
 *
 *  The user may correct typing errors in the input window with the
 *  backspace or delete keys.  (Caveats:  these keys have proven
 *  remarkably non-portable over the three decades of Unix existence,
 *  so your milage may vary.)
 *
 *  Pressing F1 prompts the user to enter a file name.  If readable,
 *  the file is read, and each line appears in the input window as though
 *  the user had typed it.
 *
 *    -> abort the reading of a command file by pressing F1 again <-
 *
 *  A note on command-script files:  they may contain comments.  A
 *  comment begins with the '#' character, and continues to the end
 *  of the line.  A low-level routine replaces the '#' with a zero-byte,
 *  which means that the comments will not appear in the input window.
 *  Lines consisting of nothing but white space will  neither be sent
 *  to the PCIx controller nor displayed in the input curses window.

 *    Usage:
 *            pcixcom [OMS-card-name [transcript-file-name]]
 *
 *      where OMS-card-name is a string of the form OmsPCIx1 for the
 *      first card in the system, OmsPCIx2 for the second, etc.
 *      (you may also enter the Linux character-special file name,
 *      for example, /dev/oms-pcix0)
 *
 *     NOTE:  case is ignored in OMS-card-name strings.  You aren't
 *            required to carefully match the upper/lower combination
 *            shown above.  For example "omspcix2" works just fine.
 *
 *     NOTE:  you can't name a transcript file without naming a PCIx ID;
 *            the first command arg is always interpreted as an
 *            OMS-card-name.
 *
 *   Normally, all text appearing in the curses windows is lost once
 *   it has scrolled away.  By running the program with the optional
 *   transcript-file-name argument, a transcript will be written to
 *   the file named.  Lines typed by the user in the input window will
 *   be prefixed with "*** " for clarity.  Otherwise, the transcript file
 *   will be written in the order that strings appear from either source.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <asm/types.h>   // need cpu-specific u32
#include <fcntl.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <string.h>
#include <errno.h>
#include <curses.h>
#include <malloc.h>
#include <unistd.h>

#include "oms-pcix.h" // needed to put driver in INTERACTIVE_MODE

// map from OMS names to linux char-special file names.  (Note:  the
//   program will operate properly if you invoke it with, for example:
//               interactive /dev/oms-pcix3
//
struct namemap {
	char    *omsname;
	char    *linuxname;
};
struct namemap namemap[] = {
        {"OmsPCIx1",  "/dev/oms-pcix0"}, {"OmsPCIx2",  "/dev/oms-pcix1"},
        {"OmsPCIx3",  "/dev/oms-pcix2"}, {"OmsPCIx4",  "/dev/oms-pcix3"},
        {"OmsPCIx5",  "/dev/oms-pcix4"}, {"OmsPCIx6",  "/dev/oms-pcix5"},
        {"OmsPCIx7",  "/dev/oms-pcix6"}, {"OmsPCIx8",  "/dev/oms-pcix7"},
        {"OmsPCIx9",  "/dev/oms-pcix8"}, {"OmsPCIx10", "/dev/oms-pcix9"},
        {"OmsPCIx11", "/dev/oms-pcix10"},{"OmsPCIx12", "/dev/oms-pcix11"},
        {"OmsPCIx13", "/dev/oms-pcix12"},{"OmsPCIx14", "/dev/oms-pcix13"},
        {"OmsPCIx15", "/dev/oms-pcix14"},{"OmsPCIx16", "/dev/oms-pcix15"},
        { NULL, NULL}
};

char  *devname;
char  *omsname;
char  *ourname;      // from main's argv[0]

FILE  *transcript_fp;    // if optional command arg asks for transcript
int    pcix_fd;          // file descriptor giving access to driver

#define COMMENT_CHAR  '#'

// function prototypes
void         get_linux_name(char *name);
void         delete_nonprinting(char *buf);
void         delete_comment(char *buf);
void         delete_trailing_whitespace(char *buf);
void         write_line_to_controller(char *buf);
int          line_is_empty(char *buf);
int          check_output_from_controller(char *buf);
void         get_driver_version_string(char *buf);
static int   check_controller_flags(char *buf);  // private
static void  help(void);

extern void  run_display(void);   // in display.c -- does not return
extern void  quit_curses(void);   // in display.c -- calls endwin()

int main(int ac, char **av)
{
	int fd;

	// make 'ourname' point to filename given in first arg.
	//  point past the rightmost '/' char in that string.
	if ((ourname = strrchr(*av, '/')) == NULL)
		ourname = *av;
	else
		ourname++;

	if (ac != 1 && ac != 2 && ac != 3) 
	{
		help();
		exit(0);
	}

	// set globals 'devname' and 'omsname' to either the defaults
	//   or based on first command-line argument.
	if (ac == 1) 
	{
		devname = namemap[0].linuxname;
		omsname = namemap[0].omsname;
	}
	else
	{
		// if the first command arg is a cry for help, give it.
		if (strstr(av[1], "help") ||
				strstr(av[1], "HELP") ||
				strstr(av[1], "Help")) 
		{
			help();
			exit(0);
		}
		// otherwise, assume first arg is a controller name.
		get_linux_name(av[1]);
		if (devname == NULL) 
		{
			fprintf(stderr, "%s:  \"%s\": name not known\n\n", ourname, av[1]);
			help();
			exit(1);
		}
	}
	
	// if there is a third arg, open it for the transcript output.
	if (ac == 3) 
	{
		if ((transcript_fp = fopen(av[2], "w")) == NULL) 
		{
			fprintf(stderr, "%s:  can't write transcript file:"
					"  %s\n", ourname, strerror(errno));
			help();
			exit(1);
		}
	}

	// open the device.
	if ((fd = open(devname, O_RDWR)) < 0) 
	{
		switch (errno) 
		{
			case ENODEV:
				fprintf(stderr, "%s:  \"%s\" (%s):  no such device\n",
						ourname, omsname, devname);
				break;
			case EBUSY:
				fprintf(stderr, "%s:  \"%s\" (%s):  device is busy\n",
					ourname, omsname, devname);
				break;
			default:
				fprintf(stderr, "%s:  open of \"%s\" (%s) failed:  "
						"%s\n", ourname, omsname, devname,
						strerror(errno));
				break;
		}
		fprintf(stderr, "\nrun \"%s help\" for a help message.\n\n",
				ourname);
//		exit(1);
	}

	// put controller in interactive mode
	if (ioctl(fd, IOCTL_OMS_ENTER_INTERACTIVE_MODE, 0)) 
	{
		if (errno == EBUSY) 
		{
			fprintf(stderr, "%s:  \"%s\" (%s):  device is busy\n",
					ourname, omsname, devname);
		}
		else
		{
			fprintf(stderr, "%s:  \"%s\" (%s):  ioctl to put\n"
					"   controller in interactive mode failed:"
					"  %s\n", ourname, omsname, devname,
					strerror(errno));
		}
//		exit(2);
	}

	// export fd to a global
	pcix_fd = fd;

	// initialize the curses display and run the program
	//  run_display() will call routines in this file as
	//  needed to interact with the PCIx controller.
	run_display();
        
	// not reached
	return 0;
}


// get_linux_name() --- look up a string in namemap[].  We declare a
//   match if EITHER the omsname OR the linuxname of a map entry
//   is the same as our argument string.  Sets globals 'devname' and
//   'omsname'.
void
get_linux_name(char *name)
{
	struct namemap *nm;

        // ignore case when comparing the strings
	for (nm = namemap; nm->omsname != NULL; nm++) {
		if (strcasecmp(name, nm->omsname) == 0)
			break;
		if (strcasecmp(name, nm->linuxname) == 0)
			break;
	}
	devname = nm->linuxname;
	omsname = nm->omsname;
}

// delete_nonprinting() --- replace the first non-printing char with '\0'
void
delete_nonprinting(char *buf)
{
	while (*buf && (*buf >= ' ' && *buf <= '~'))
		buf++;
	*buf = 0;
}

// delete_comment() --- search string for COMMENT_CHAR.  If found,
//     replace it with '\0'
void
delete_comment(char *buf)
{
        while (*buf && *buf != COMMENT_CHAR)
                buf++;
        *buf = 0;
}

// line_is_empty() --- predicate returning TRUE if a buffer contains
//   nothing but non-printable characters
int
line_is_empty(char *buf)
{
        while (*buf) {
                if (*buf > ' ' && *buf <= '~')
                        return 0;   // found a printable character
                buf++;
        }
        return 1;  // true.  nothing, or nothing but white space
}


// delete_trailing_whitespace() --- delete any trailing whitespace
void
delete_trailing_whitespace(char *buf)
{
        char *p;

        for (p = buf; *p; p++)
                ; // nothing
        p--;
        while (p >= buf) {
                if (*p == ' ' || *p == '\t')
                        *p-- = 0;
                else
                        return;
        }
}

// write_line_to_controller() --- delete any trailing newline and
//    write the line to the controller.  If write(2) returns 0, try
//    again up to a fairly large number of retry attempts.
//
//  This function called from run_display() in display.c to send
//  a line of user input to the controller (and transcript file).
//
//  If we're writing a transcript, append this line to that file
void
write_line_to_controller(char *buf)
{
        int write_tries = 500;
        int n;
        
        delete_nonprinting(buf);   // strip any non-printing chars

        if (buf[0] == 0)
                return;

        if (transcript_fp) {
		// pre-pend "*** " for identifying string direction
                fprintf(transcript_fp, "*** %s\n", buf);
                fflush(transcript_fp);
        }

        while (--write_tries >= 0) {
                n = write(pcix_fd, buf, strlen(buf));
                if (n < 0) {
                        if (errno == EINTR)
                                continue;
                        quit_curses();
                        fprintf(stderr, "%s:  write(2) failed on %s (%s):  "
                                "%s\n", ourname, omsname, devname,
                                strerror(errno));
                        exit(1);
                }
                if (n > 0)
                        return;   // driver only writes ALL if it, or none.

                // if (n == 0), driver has already delayed a short interval
                //    so that controller firmware can consume data already
                //    in its command buffer.  Retry the write(2).
        }
        quit_curses();
        fprintf(stderr, "%s:  failed to write to controller:  too many"
                " retries.\n", ourname);
        exit(1);
}

// check_output_from_controller() --- see if the controller has anything
//    in its response buffer, or if any flags are set.  In the former
//    case, read(2) the response buffer, and in the latter case,
//    construct a message indicating the flag event(s).
//
//  If we're writing a transcript, put any controller-generated strings
//   in that file.
//
//  This routine is called from run_display() in display.c to poll
//  the PCIx controller for anything that should be displayed.
//
//  Use select(2) with short timeouts so that code in display.c can
//  interact with the user.
//
//  Returns 1 if there is a line, else 0.
int
check_output_from_controller(char *buf)
{
	struct timeval  tv;
	fd_set          read_fds;
	int             n = 0;
	int             rv;
        
	FD_ZERO(&read_fds);
	FD_SET(pcix_fd, &read_fds);
	tv.tv_sec = 0;
	tv.tv_usec = 1000;
	rv = select(pcix_fd + 1, &read_fds, NULL, NULL, &tv);
	if (rv < 0 && errno != EINTR) 
	{
		quit_curses();
		fprintf(stderr, "%s:  select(2) failed on %s:  %s\n",
                        ourname, omsname, strerror(errno));
		exit(1);
	}
	if (rv > 0) 
	{
		// read the data.  Watch for EINTR.
		while (1) 
		{
			n = read(pcix_fd, buf, 1024);  // better be big enuf!
			if (n < 0) 
			{
				if (errno == EINTR)
					continue;
				quit_curses();
				fprintf(stderr, "%s:  read(2) failed on %s:  %s\n",
                ourname, omsname, strerror(errno));
				exit(1);
			}
			if (n == 0)
				break;
			buf[n] = 0;
			delete_nonprinting(buf);
			delete_trailing_whitespace(buf);
			if (transcript_fp) 
			{
				fprintf(transcript_fp, "%s\n", buf);
				fflush(transcript_fp);
			}
			return 1;  // we put data in caller's buffer
		}
	}
	// if we fall through to here, there is no text response
	// check the flags
	return check_controller_flags(buf);
}

void
capture_position_data(char *buf)
{
	struct oms_ioctl  ioc;
	static __u32      pos[16];
	char              tmp[128];

	// we will append things to buffer.  Start with it empty.
	*buf = 0;

	// can only do position capture outside of interactive mode.
//	ioctl(pcix_fd, IOCTL_OMS_EXIT_INTERACTIVE_MODE, 0);


	memset(&ioc, 0, sizeof(ioc));
	ioc.response = (char *)&pos;
	ioc.response_len = sizeof(__u32) * 16;
	if (ioctl(pcix_fd, IOCTL_OMS_READ_AXIS_POSITIONS, &ioc)) 
	{
		quit_curses();
		fprintf(stderr, "%s:  IOCTL_OMS_READ_AXIS_POSITIONS:  %s\n",
                        ourname, strerror(errno));
		exit(1);
	}
	
	sprintf( tmp, "X Motor: %d, Encoder: %d\n", pos[0], pos[8] );
	strcat( buf, tmp );
	sprintf( tmp, "Y Motor: %d, Encoder: %d\n", pos[1], pos[9] );
	strcat( buf, tmp );
	sprintf( tmp, "Z Motor: %d, Encoder: %d\n", pos[2], pos[10] );
	strcat( buf, tmp );
	sprintf( tmp, "T Motor: %d, Encoder: %d\n", pos[3], pos[11] );
	strcat( buf, tmp );
	sprintf( tmp, "U Motor: %d, Encoder: %d\n", pos[4], pos[12] );
	strcat( buf, tmp );
	sprintf( tmp, "V Motor: %d, Encoder: %d\n", pos[5], pos[13] );
	strcat( buf, tmp );
	sprintf( tmp, "R Motor: %d, Encoder: %d\n", pos[6], pos[14] );
	strcat( buf, tmp );
	sprintf( tmp, "S Motor: %d, Encoder: %d\n", pos[7], pos[15] );
	strcat( buf, tmp );

	// put controller back in interactive mode
//	ioctl(pcix_fd, IOCTL_OMS_ENTER_INTERACTIVE_MODE, 0);
}


//**********************************
// controller status flag checking

// we don't use the shared library (just the device driver), so
//  we don't have OMS_X_AXIS, et. al. defined.
#ifndef OMS_X_AXIS
#define OMS_X_AXIS  (1 << 0)
#define OMS_Y_AXIS  (1 << 1)
#define OMS_Z_AXIS  (1 << 2)
#define OMS_T_AXIS  (1 << 3)
#define OMS_U_AXIS  (1 << 4)
#define OMS_V_AXIS  (1 << 5)
#define OMS_R_AXIS  (1 << 6)
#define OMS_S_AXIS  (1 << 7)
#endif


static void
append_axis_flags_string(__u8 flags, char *buf)
{
        if (flags & OMS_X_AXIS) strcat(buf, " X");
        if (flags & OMS_Y_AXIS) strcat(buf, " Y");
        if (flags & OMS_Z_AXIS) strcat(buf, " Z");
        if (flags & OMS_T_AXIS) strcat(buf, " T");
        if (flags & OMS_U_AXIS) strcat(buf, " U");
        if (flags & OMS_V_AXIS) strcat(buf, " V");
        if (flags & OMS_R_AXIS) strcat(buf, " R");
        if (flags & OMS_S_AXIS) strcat(buf, " S");
}

// check_controller_flags() --- read the driver's copy of the controller
//    flags, and construct a string describing what is set.  Clear
//    the driver's copy of the flags.
//
//   Returns 0 if no flag data is available, 1 if we constructed a string.
//
//  (In theory, the string could be quite long if all the flags are set
//   at once.  Probably won't happen ....)
static int
check_controller_flags(char *buf)

{
        struct oms_ioctl  ioc;
        static __u8       flags;

        // we will append things to buffer.  Start with it empty.
        *buf = 0;

				flags = 0;
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *)&flags;
        ioc.response_len = sizeof(__u8);
        if (ioctl(pcix_fd, IOCTL_OMS_READ_STATUS_FLAGS, &ioc)) {
                quit_curses();
                fprintf(stderr, "%s:  IOCTL_OMS_READ_STATUS_FLAGS:  %s\n",
                        ourname, strerror(errno));
                exit(1);
        }
        if ( flags & CMDERR_BIT ) {
          char tmp[32];
             
          sprintf(tmp, "CMD ERR FLAGS:  %x", CMDERR_BIT);
          strcat(buf, tmp);

	        // clear the driver copy so we won't see them next time
	        memset(&ioc, 0, sizeof(ioc));
	        ioc.command = (char *) &flags;
	        ioc.command_len = sizeof(__u8);
	        if (ioctl(pcix_fd, IOCTL_OMS_CLEAR_STATUS_FLAGS, &ioc)) {
	                quit_curses();
	                fprintf(stderr, "%s:  IOCTL_OMS_CLEAR_STATUS_FLAGS:"
	                        "  %s\n", ourname, strerror(errno));
	                exit(1);
	        }
				}

				flags = 0;
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *)&flags;
        ioc.response_len = sizeof(__u8);
        if (ioctl(pcix_fd, IOCTL_OMS_READ_DONE_FLAGS, &ioc)) 
				{
                quit_curses();
                fprintf(stderr, "%s:  IOCTL_OMS_READ_DONE_FLAGS:  %s\n",
                        ourname, strerror(errno));
                exit(1);
        }
        // clear the driver copy so we won't see them next time
        if ( flags ) {
          strcat(buf, "DONE FLAGS:  ");
          append_axis_flags_string(flags, buf);
          strcat(buf, "  ");

	        memset(&ioc, 0, sizeof(ioc));
	        ioc.command = (char *) &flags;
	        ioc.command_len = sizeof(__u8);
	        if (ioctl(pcix_fd, IOCTL_OMS_CLEAR_DONE_FLAGS, &ioc)) {
	                quit_curses();
	                fprintf(stderr, "%s:  IOCTL_OMS_DONE_STATUS_FLAGS:"
	                        "  %s\n", ourname, strerror(errno));
	                exit(1);
	        }
				}

				flags = 0;
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *)&flags;
        ioc.response_len = sizeof(__u8);
        if (ioctl(pcix_fd, IOCTL_OMS_READ_PLIMIT_FLAGS, &ioc)) 
				{
                quit_curses();
                fprintf(stderr, "%s:  IOCTL_OMS_READ_PLIMIT_FLAGS:  %s\n",
                        ourname, strerror(errno));
                exit(1);
        }
        // clear the driver copy so we won't see them next time
        if ( flags ) {
          strcat(buf, "POS LIMIT FLAGS: ");
          append_axis_flags_string( flags, buf);
          strcat(buf, "  ");

	        memset(&ioc, 0, sizeof(ioc));
	        ioc.command = (char *) &flags;
	        ioc.command_len = sizeof(__u8);
	        if (ioctl(pcix_fd, IOCTL_OMS_CLEAR_PLIMIT_FLAGS, &ioc)) {
	                quit_curses();
	                fprintf(stderr, "%s:  IOCTL_OMS_PLIMIT_STATUS_FLAGS:"
	                        "  %s\n", ourname, strerror(errno));
	                exit(1);
	        }
				}

				flags = 0;
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *)&flags;
        ioc.response_len = sizeof(__u8);
        if (ioctl(pcix_fd, IOCTL_OMS_READ_NLIMIT_FLAGS, &ioc)) 
				{
                quit_curses();
                fprintf(stderr, "%s:  IOCTL_OMS_READ_NLIMIT_FLAGS:  %s\n",
                        ourname, strerror(errno));
                exit(1);
        }
        // clear the driver copy so we won't see them next time
        if ( flags ) {
          strcat(buf, "NEG LIMIT FLAGS: ");
          append_axis_flags_string(flags, buf);
          strcat(buf, "  ");

	        memset(&ioc, 0, sizeof(ioc));
	        ioc.command = (char *) &flags;
	        ioc.command_len = sizeof(__u8);
	        if (ioctl(pcix_fd, IOCTL_OMS_CLEAR_NLIMIT_FLAGS, &ioc)) {
	                quit_curses();
	                fprintf(stderr, "%s:  IOCTL_OMS_NLIMIT_STATUS_FLAGS:"
	                        "  %s\n", ourname, strerror(errno));
	                exit(1);
	        }
				}

				flags = 0;
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = (char *)&flags;
        ioc.response_len = sizeof(__u8);
        if (ioctl(pcix_fd, IOCTL_OMS_READ_SLIP_FLAGS, &ioc)) 
				{
                quit_curses();
                fprintf(stderr, "%s:  IOCTL_OMS_READ_SLIP_FLAGS:  %s\n",
                        ourname, strerror(errno));
                exit(1);
        }
        // clear the driver copy so we won't see them next time
        if ( flags ) {
          strcat(buf, "SLIP FLAGS: ");
          append_axis_flags_string(flags, buf);
          strcat(buf, "  ");

	        memset(&ioc, 0, sizeof(ioc));
	        ioc.command = (char *) &flags;
	        ioc.command_len = sizeof(__u8);
	        if (ioctl(pcix_fd, IOCTL_OMS_CLEAR_SLIP_FLAGS, &ioc)) {
	                quit_curses();
	                fprintf(stderr, "%s:  IOCTL_OMS_CLEAR_SLIP_FLAGS:"
	                        "  %s\n", ourname, strerror(errno));
	                exit(1);
	        }
				}

        // we may have some trailing whitespace.  get rid of it
        delete_trailing_whitespace(buf);
        // append to transcript, if any
        if (transcript_fp) {
                fprintf(transcript_fp, "%s\n", buf);
                fflush(transcript_fp);
        }
        return strlen(buf);
}                

// get_driver_version_string() --- retrieve the version string from
//   the device driver.
void
get_driver_version_string(char *buf)
{
        struct oms_ioctl ioc;
        
        memset(&ioc, 0, sizeof(ioc));
        ioc.response = buf;
        ioc.response_len = 256;  // we hope not to overflow caller's buffer
        if (ioctl(pcix_fd, IOCTL_OMS_GET_VERSION, &ioc)) {
                // put an error message in caller's buffer
                sprintf(buf, "IOCTL_OMS_GET_VERSION failed:  %s",
                        strerror(errno));
        }
        if (transcript_fp) {
                fprintf(transcript_fp, "%s\n", buf);
                fflush(transcript_fp);
        }
}
        
// the help message
static char *help_msg[] = {
        "\n",
        "OMS Interactive program for PCIx\n\n",
        "Usage:  %s [PCIx-name [transcript-file]]\n",
        "   where PCIx-name is of form \"OmsPCIx1\" (case ignored)\n",
        "\n",
        "Two windows will be presented --- you type in the lower\n",
        "and output from the PCIx appears in the upper.\n",
        "\n",
        "Press F1 to be prompted for a file name.  Commands will be\n",
        "read from the file as though you had typed them.\n",
        "\n",
        "To abort command file processing, press F1 again.\n",
        "\n",
        "Note:  command files may have comments beginning with '#'\n",
        "\n",
        "Press <CTRL>C to end your interactive session.\n",
        "\n",
        NULL
};

// help() --- print the help message
static void
help(void)
{
        char **s;

        for (s = help_msg; *s; s++) {
                printf(*s, ourname);
        }
}
