/*
 *  oms-core.c --- Linux device driver for Oregon Micro Systems
 *    motion controllers, 1 of at least 2. This file contains the
 *		generic IOCTL and read/write/poll code. Hardware-specific code
 *		will be located elsewhere.
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 *
 *  Theory of queued ioctl operation
 *     This driver is built around the most common operation:
 *         -->> putting a string into the controller command buffer, and,
 *         -->>  optionally waiting for a response
 *
 *     Because the driver must support multiple simultaneous opens,
 *     each of which may conduct controller operations, the operations
 *     are serialized by placing them into a simple queue, which is
 *     consumed one at a time.  (There are some exceptions, e.g.,
 *     an operation that reads the current number of bytes available
 *     in the controller's command buffer.)
 *
 *     Virtually all controller commands use a single ioctl data
 *     structure, described in oms-ioctl.h:
 *
 *           struct oms_ioctl {   // user/kernel ioctl API definition
 *                   unsigned char *command;       // IN
 *                   unsigned int   command_len;   // IN
 *                   unsigned char *response;      // OUT
 *                   unsigned int   response_len;  // IN and OUT
 *           };
 *
 *     This user-space structure is copied into a newly allocated
 *     instance of struct omsdrv_ioctl, which contains other structure
 *     elements (pointers, flags, etc.).  If a user-space command string
 *     is present, space for it is allocated and the command string is
 *     copied into kernel space.  If a response string is expected, then
 *     kernel space is pre-allocated.  The structure is put at the tail
 *     of the queue.
 *
 *     Once queued, the user-thread executing the ioctl function sleeps
 *     until awakened, or until a timer expires.  (Note:  due to the
 *     essential non-restartability of most controller operations, the
 *     sleep is UNINTERRUPTIBLE --- we rely on the timer to prevent
 *     indefinite hangs.)
 *
 *     When awakened, the user-thread code executing the ioctl system
 *     call completes the operation, which is frequently just copying
 *     out the oms_ioctl structure, along with any response data,
 *     back to the user program.
 *
 *  INTERACTIVE_MODE --- the driver enforces a single-open policy
 *     for interactive mode:  you can't put the driver into this mode
 *     if there is another active open, and you can't open the driver
 *     if it's in interactive mode.
 *
 *     For interactive mode, three new entry points are supported:
 *       read() --- read text responses from the controller
 *       write() --- write to the command buffer of the controller
 *       poll() --- supports the select(2) syscall so that the
 *                  application can determine when data is available
 *                  for reading.
 *
 *       The read and write routines are non-blocking.  Read will
 *       return immediately if there are no waiting text responses.
 *       Write will return immediately if there is insufficient room
 *       in the controller's command buffer for the entire command string.
 *
 *       Text response data is delivered to the driver via the normal
 *       interrupt mechanism.  The IRQ handler notices that the driver
 *       is in interactive mode and allocates a simple data structure
 *       to hold the response data, which is then retrieved from the
 *       controller and held pending a subsequent read.
 *       (This contrasts to the normal-mode operation, where text is
 *       retrieved from the controller only if there is an IOCTL
 *       struct waiting for it .... if there is none, the controller's
 *       response is discarded.)
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/config.h>
#include <linux/version.h>

#ifdef CONFIG_MODVERSIONS
#ifndef MODVERSIONS
#define MODVERSIONS
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,40)
#include <linux/modversions.h>
#endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/kmod.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <linux/byteorder/generic.h>

#include "oms-core.h" // shared with user-space programs:  defines IOCTLs

MODULE_AUTHOR("Oregon Micro Systems, Inc.");
//MODULE_DESCRIPTION found in hardware-specific source
MODULE_LICENSE("GPL");


//  ********** module parameters **********
int DEBUGLEVEL = 0;
MODULE_PARM(DEBUGLEVEL, "i");
MODULE_PARM_DESC(DEBUGLEVEL, "Debug output level");

int MAJORNUMBER = OMS_MAJOR_NUMBER;
MODULE_PARM(MAJORNUMBER, "i");
MODULE_PARM_DESC(MAJORNUMBER, "Major number");

int ONE_CPU = TRUE;
MODULE_PARM(ONE_CPU, "i");
MODULE_PARM_DESC(ONE_CPU, "Single CPU environment");

int MAX_CARDS = 16;
MODULE_PARM(MAX_CARDS, "i");
MODULE_PARM_DESC(MAX_CARDS, "Max supported cards (default=16)");

// This driver is accompanied by a script that will create the proper
//  char-special files (aka "nodes") under /dev.  This helper program
//  is useful, as we're using dynamic major number allocation.  (For now.)
char* NODE_CREATION_SCRIPT = "/opt/oms/bin/create_nodes";
MODULE_PARM(NODE_CREATION_SCRIPT, "s");
MODULE_PARM_DESC(NODE_CREATION_SCRIPT, "Path to node creation script");


module_init(oms_install_module);      // hook for insmod
module_exit(oms_remove_module);   		// hook for insmod


// ********* data declarations *****************
kmem_cache_t* oms_slab;
LIST_HEAD(oms_cardlist);
unsigned long flags;						//for locking, driver global

extern char* me;								//from card-specific source
extern char version_string[];		//from card-specific source
// the OMS device ID will be some value + a small integer set by DIP switches
extern int PCI_DEVICE_ID_OMS_BASE;

#if OMS_DEBUG_SPINLOCKS
	atomic_t   oms_islocked;
#endif


// pci_device_id --- specify to PCI driver the set of vendor/device IDs
//     that we know about.  The PCI driver will call our "probe" routine
//     for each device matching one of our descriptions.
extern struct pci_device_id oms_pci_tbl[] __devinitdata;
MODULE_DEVICE_TABLE(pci, oms_pci_tbl);


// pci_driver structure --- register with the PCI driver.  It will call
//   our "probe" routine for each PCI device matching our table.
extern struct pci_driver oms_driver;


// file_operations table for registering as a char driver
struct file_operations oms_fileops = 
{
	.owner   = THIS_MODULE,   // automatic handing of module refcounts
	.open    = oms_fop_open,
	.release = oms_fop_close,
	.ioctl   = oms_fop_ioctl,
	.read    = oms_fop_read,     // supported only in INTERACTIVE_MODE
	.write   = oms_fop_write,    // supported only in INTERACTIVE_MODE
	.poll    = oms_fop_poll      // supported only in INTERACTIVE_MODE
};

DECLARE_TASKLET( oms_tasklet, oms_tasklet_ioctl, 0 );

extern char* node_device;


//***************************************************************
//   beginning of code





//****************************************************************************
// oms_install_module() -- called when the module is loaded, but before
//   it is available for user programs.  Here, we register as a PCI
//   driver, which will indirectly call our "probe" routine for each
//   PCI device matching the specs in oms_pci_tbl[].  Our probe routine
//   is oms_install_one().  If no cards were found, the module fails
//   to load.  (Fix later if hot-swap CompactPCI is required.)  Otherwise,
//   we register as a character driver.  During this registration, we
//   first attempt to use our default major number.  Should this fail,
//   we fall back on using a dynamically assigned major number.  In either
//   case, we try to run a user-mode shell script that ensures creation
//   of the char-special files (aka "nodes") in /dev.
//
//   We also create a "slab allocator" that we will use for the many
//   allocations of 'struct omsdrv_ioctl' necessary for queued ioctl
//   operations.  (This is a very fast and efficient means of allocating
//   specific objects.)
int __init oms_install_module(void)
{
	int retval;
	
	// register as a PCI driver.  The upper layer will call our "probe"
	//  routine for each card that should match our specifications.
	retval = pci_module_init( &oms_driver );
	if ( !retval )
		printk( KERN_INFO "%s:  module PCI init successful\n", me );
	else
	{
		printk( KERN_ERR "%s:  module failed to load:  no cards found\n", me );
		return retval;
	}
	
	// create a slab allocator for the frequently needed ioctl ops
	oms_slab = kmem_cache_create( me, sizeof(struct omsdrv_ioctl), 0,
								  SLAB_HWCACHE_ALIGN, NULL, NULL );
	if ( oms_slab == NULL ) 
	{
		printk( KERN_ERR "%s:  kmem_cache_create failed\n", me );
		pci_unregister_driver( &oms_driver );
		_LOCK_( flags );
		oms_release_resources(NULL);
		UN_LOCK_( flags );
		return -ENODEV;
	}

	// we found and initialized some cards.  Register as a chrdev.
	if ( register_chrdev( MAJORNUMBER, me, &oms_fileops ) < 0 ) 
	{
		printk( KERN_ERR "%s:  failed to register as char-major %d\n", me, MAJORNUMBER );
		pci_unregister_driver( &oms_driver );
		_LOCK_( flags );
		oms_release_resources(NULL);
		UN_LOCK_( flags );
		return -ENODEV;
	}
	
	oms_create_nodes();

	printk( KERN_INFO "%s:  module loaded successfully\n", me );	
	return retval;
}
//****************************************************************************


//****************************************************************************
// oms_remove_module() ---  called when our module is unloaded.  As
//   with module initialization, most of the work is done indirectly.
//   Here, we call pci_unregister_driver(), which will call our
//   'remove' function (oms_remove_one()) for each card we've claimed.
//   We double check that all of the cards have indeed been removed.
//   (Early testing showed no actual need for this.)  We also
//   unregister as a character driver and release our slab allocator.
void __exit oms_remove_module(void)
{
  tasklet_kill( &oms_tasklet );

	if ( MAJORNUMBER > 0 )
	{
		(void)unregister_chrdev( MAJORNUMBER, me );
	}
	
	pci_unregister_driver( &oms_driver );
	_LOCK_( flags );
	oms_release_resources(NULL);   // double check resource release
	UN_LOCK_( flags );
	printk( KERN_INFO "%s:  module cleanup\n", me );
	
	if ( !list_empty( &oms_cardlist ) )
		printk( KERN_ERR "%s:  non-empty card list\n", me );
	
	// delete our slab allocator
	kmem_cache_destroy( oms_slab );
}
//****************************************************************************


//****************************************************************************
// oms_install_one ---- called by the driver "probe" routine for each
//    PCI device found matching our pci_device_id table.
int __devinit oms_install_one( struct pci_dev* pdev,
                               const struct pci_device_id* ent )
{
	int             	minor;
	struct oms_card*	card;
	struct oms_card*  temp;
	struct list_head* ptr;

	ASSERT_UNLOCKED();
	
	// First, derive the minor number from the DIP switch settings
	//  that contributes to the subsystem_device value
	if ( pdev->subsystem_device < PCI_DEVICE_ID_OMS_BASE ) 
	{
		minor = 16;   // fake an error --- see below
	}
	else
	{
		// the card firmware translates DIP switch value 0 to 1
		//    Subtract 1 to get to minor range 0 to (MAX_CARDS -1)
		minor = pdev->subsystem_device - PCI_DEVICE_ID_OMS_BASE - 1;
	}
	if (minor >= 16)	//never support more than 16 cards at a time
	{
		printk( KERN_ERR "%s:  invalid subsystem_device ID %04x\n", me, pdev->subsystem_device );
		return -ENODEV;
	}
	
	// from here on, we're under the 'big' spinlock
	_LOCK_( flags );

	// verify that we don't already have a card with this subsystem_device ID
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		temp = list_entry( ptr, struct oms_card, listhead );
		if ( temp->minor == minor )
		{
			UN_LOCK_(flags);
			printk( KERN_ERR "%s:  duplicate subsystem_device ID %04x\n", me, pdev->subsystem_device );
			printk( KERN_ERR "%s:  (minor %d).  Adjust DIP switches so all cards are unique.\n", me, minor );
			return -EBUSY;
		}
	}

	if ( pci_enable_device( pdev ) != 0 ) 
	{
		UN_LOCK_(flags);
		return -ENODEV;
	}

	card = kmalloc( sizeof(struct oms_card), GFP_ATOMIC );	//cannot allow sleep here
	memset(card, 0, sizeof(struct oms_card));
	card->pdev						= pdev;
	card->minor						= minor;
	card->irq     				= -1;
	card->blocking				= 0;
	card->response_ready	= 0;
	INIT_LIST_HEAD(&card->listhead);
	INIT_LIST_HEAD(&card->ioctl_list);
	init_waitqueue_head(&card->poll_waitq);
	list_add( &card->listhead, &oms_cardlist );

	if ( _oms_init_memory( pdev, card ) != 0 || 
			 _oms_attach_irq( pdev, card ) != 0 )
	{
		oms_release_resources( card );
		UN_LOCK_(flags);
		return -ENODEV;
	}

	tasklet_schedule( &oms_tasklet );  // get things moving

	UN_LOCK_(flags);
	printk( KERN_INFO "%s:  device initialized (minor %d)\n", me, minor );
	return 0;
}
//****************************************************************************


//****************************************************************************
// oms_remove_one() --- normally called from pci_unregister_driver().
//   Release all resources associated with an oms card we've previously
//   claimed.
void oms_remove_one( struct pci_dev *pdev )
{
	struct oms_card*	card;
  struct list_head* ptr;

	ASSERT_UNLOCKED();

	_LOCK_(flags);

	// search for the card matching the pdev by bus/slot number
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if (card->pdev->bus->number == pdev->bus->number && card->pdev->devfn == pdev->devfn) 
		{
			oms_release_resources( card );
			UN_LOCK_(flags);
			return;
		}
	}
	printk(KERN_WARNING "%s:  remove_one failed to find bus %d, slot %d\n", me, pdev->bus->number, PCI_SLOT(pdev->devfn));
	UN_LOCK_(flags);
}
//****************************************************************************


//****************************************************************************
// oms_release_resources() --- release all resources used by one, or
//   all, cards.  If arg 'card' is NULL release for all
//   else only the specified card.
//
//  Note:  spinlock oms_biglock is LOCKED on entry.
void oms_release_resources( struct oms_card* card )
{
  struct list_head* ptr;
	
	ASSERT_LOCKED();

	if ( card == NULL )
	{	
		for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
		{
			struct oms_card* card = list_entry( ptr, struct oms_card, listhead );
			oms_release_resources( card );
		}
		return;
	}
	
	_oms_release_resources( card );
	list_del( &card->listhead );
	for ( ptr = card->ioctl_list.next; ptr != &(card->ioctl_list); ptr = ptr->next )
	{
		struct omsdrv_ioctl* sioctl = list_entry( ptr, struct omsdrv_ioctl, listhead );
		oms_release_ioctl( sioctl );
	}
	kfree( card );
}
//****************************************************************************


//****************************************************************************
// run a user-mode program to ensure that the correct nodes in /dev are
//    present.  The program is a shell script that is part of the package.
void oms_create_nodes(void)
{
	char *av[5];
	char  major_buf[8];
	char  max_buf[8];
	char *env[] = {NULL};
	
	sprintf( major_buf, "%d", MAJORNUMBER );
	sprintf( max_buf, "%d", MAX_CARDS );
	av[0] = NODE_CREATION_SCRIPT;
	av[1] = node_device;
	av[2] = major_buf;
	av[3] = max_buf;
	av[4] = NULL;
	
	(void)call_usermodehelper( NODE_CREATION_SCRIPT, av, env, 1 );
}
//****************************************************************************




//****************************************************************************
// oms_new_ioctl() --- allocate an instance of struct omsdrv_ioctl from
//   our private slab allocator, zero it, and return it.
struct omsdrv_ioctl* oms_new_ioctl( struct oms_card* card )
{
	struct omsdrv_ioctl* sioctl;

	if ( card == NULL ) 
	{
		printk(KERN_ERR "%s:  oms_new_ioctl:  NULL card\n", me);
	}

	sioctl = kmem_cache_alloc( oms_slab, GFP_KERNEL );
	memset( sioctl, 0, sizeof(struct omsdrv_ioctl) );
	sioctl->card = card;

	return sioctl;
}
//****************************************************************************


//****************************************************************************
// oms_release_ioctl() --- free the resources associated with one or
//   more instances of 'struct omsdrv_ioctl'.  Dequeue it if it is
//   still enqueued
void oms_release_ioctl( struct omsdrv_ioctl* sioctl )
{
	if ( sioctl )
	{
		// not all ioctls get added to a list
		if ( sioctl->listhead.next )
		{
			list_del( &(sioctl->listhead) );
		}
	
		if ( sioctl->buffers.command ) 
		{
			kfree( sioctl->buffers.command );
		}
		if ( sioctl->buffers.response ) 
		{
			kfree( sioctl->buffers.response );
		}
		if ( sioctl->task_asleep ) 
		{
			printk(KERN_WARNING "%s:  task asleep on freed ioctl\n", me );
		}
	
		kmem_cache_free( oms_slab, sioctl );  // return struct to slab
	}
}
//****************************************************************************


//****************************************************************************
// oms_fetch_user_ioctl_struct() --- called from the oms_fop_ioctl() function
//    (a file_operations routine running the user thread).   Fetch the
//    user data into kernel space according to parameter args.  Also,
//    if indicated by parameter args, pre-allocate space for a response
//
//    If there is an error, the ios structure is released and an error
//    code is returned suitable for return to the user.
//
//   (we must NOT be locked, as one should never access user space under
//    a spinlock --- there might be a page fault during transfer)
int oms_fetch_user_ioctl_struct(	struct omsdrv_ioctl* pioctl,
                                  int command_parm,
                                  int response_parm,
                                  void* user_arg )
{
	void* ptmp;

	ASSERT_UNLOCKED();

	if ( pioctl == NULL || user_arg == NULL )
	{
		return -EINVAL;
	}

	if ( copy_from_user( &pioctl->buffers, user_arg, sizeof(struct oms_ioctl) ) )
	{
		oms_release_ioctl( pioctl );
		return -EFAULT;
	}
	
	if ( command_parm ) 
	{
		// we must try to copy the command string into kernel
		//  space.  If command_parm is greater than 1, then
		//  we validate by checking to see if the command_len
		//  is a multiple of command_parm.
		if ( pioctl->buffers.command_len > 4096 ) 
		{
			oms_release_ioctl( pioctl );
			return -EINVAL;
		}
		if ( command_parm > 1 && (pioctl->buffers.command_len % command_parm) != 0 ) 
		{
			oms_release_ioctl( pioctl );
			return -EINVAL;
		}
		ptmp = pioctl->buffers.command;
		pioctl->buffers.command = kmalloc( pioctl->buffers.command_len, GFP_KERNEL );
		if ( copy_from_user( pioctl->buffers.command, ptmp, pioctl->buffers.command_len ) ) 
		{
			oms_release_ioctl( pioctl );
			return -EFAULT;
		}
	}

	if ( response_parm ) 
	{
		// pre-allocate space for a response, with length constraint checking
		if ( pioctl->buffers.response_len == 0 || pioctl->buffers.response_len > 4096 )
		{
			oms_release_ioctl( pioctl );
			return -EINVAL;
		}

		if ( response_parm > 1 && (pioctl->buffers.response_len % response_parm) != 0 )
		{
			oms_release_ioctl( pioctl );
			return -EINVAL;
		}

		// make sure user allocated space for the response
		if ( pioctl->buffers.response == NULL )
		{
			oms_release_ioctl( pioctl );
			return -EINVAL;
		}

		pioctl->buffers.response = kmalloc( pioctl->buffers.response_len + 1, GFP_KERNEL );
		pioctl->p_response = 0;  // will increment as data arrives
	}

	return 0;
}
//****************************************************************************


//****************************************************************************
// oms_sleep() --- delay for a specified number of jiffies.
//    We simply sleep on a wait queue that no one ever awakens.
void oms_sleep( int jifs )
{
	wait_queue_head_t wq;
	init_waitqueue_head( &wq );
	sleep_on_timeout( &wq, jifs );
}
//****************************************************************************


//****************************************************************************
// oms_fop_open() --- file_operations function.  Verify that a card
//   exists corresponding to the minor number, and bump the count
//   of current opens.
//   Addition:  suport for INTERACTIVE_MODE --- refuse to allow open
//      if card is already in INTERACTIVE mode.  (We enforce exclusive
//      open for interactive mode here, and in the ioctl that turns on
//      that mode.)
int oms_fop_open( struct inode* pinode, struct file* pfile )
{
	int               minor;
	struct oms_card*  card;
	struct list_head* ptr;
	int               rv = 0;

	ASSERT_UNLOCKED();

	minor = MINOR( pinode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;

	_LOCK_(flags);
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		rv = -ENODEV;
	}
	else if ( card->comm_mode == INTERACTIVE_MODE ) 
	{
		printk(KERN_INFO "%s:  open of card in INTERACTIVE_MODE.  EBUSY\n", me);
		rv = -EBUSY;
	}
	else 
	{
		printk( KERN_INFO "%s:  opened %s%d\n", me, me, minor );
		++(card->open_count);
	}
  UN_LOCK_(flags);

  return rv;
}
//****************************************************************************


//****************************************************************************
// oms_fop_close() --- file_operations function.  Decrement count of opens.
//   If open_count falls to zero, take card out of INTERACTIVE_MODE so
//   that other programs may open and use the device.
int oms_fop_close( struct inode* pinode, struct file* pfile )
{
	int               minor;
	struct oms_card*  card;
	struct list_head* ptr;
	int               rv = 0;
	
	ASSERT_UNLOCKED();
	
	minor = MINOR( pinode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;   // should never happen

	_LOCK_(flags);
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		rv = -ENODEV;     // should never happen
	} 
	else 
	{
		card->blocking = 0;
		--(card->open_count);
		if ( card->open_count == 0 ) 
		{
			// no remaining opens.  take out of interactive mode
			//   (if applicable)
			card->comm_mode = STRING_MODE;
		}
	}

	printk( KERN_INFO "%s:  closed %s%d\n", me, me, minor );

	UN_LOCK_(flags);

	return rv;
}
//****************************************************************************


//****************************************************************************
// oms_fop_read() --- valid ONLY if card->comm_mode is INTERACTIVE_MODE.
//
// NON-blocking read.  If there's nothing to read, returns 0 immediately.
//     Use select(2) or poll(2) to avoid busy looping.  Each time we're
//     called we grab ALL data from the card into a temporary buffer, and
//     copy out what will fit in the user's buffer.  Any residue in
//     temporary buffer is silently discarded.  This is semi-okay, as
//     INTERACTIVE_MODE is intended only for use by OMS user programs,
//     which can be developed to allocate a large enough buffer.
int oms_fop_read( struct file* pfile,
                  char* userbuf,
                  size_t usersiz,
                  loff_t* ppos )
{
	struct oms_card*  card;
	int               minor;
	int               rv = 0;
	char*			        tempbuf;
	struct list_head* ptr;
	
	ASSERT_UNLOCKED();

	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;

	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}
	
	// ensure card is in interactive mode
	if ( card->comm_mode != INTERACTIVE_MODE )
		return -ENODEV;
        
	// grab temporary buffer and copy data in before locking
	if ( usersiz > INTERACTIVE_BUFFER_LEN )
		usersiz = INTERACTIVE_BUFFER_LEN;
	tempbuf = kmalloc( usersiz, GFP_KERNEL );

	_LOCK_(flags);
	// perform card-specific read work
	rv = _oms_fop_read( card, tempbuf, usersiz );
	UN_LOCK_(flags);

	if (copy_to_user(userbuf, tempbuf, rv))
		rv = -EFAULT;

	kfree( tempbuf );

	return rv;
}
//****************************************************************************


//****************************************************************************
// oms_fop_write --- valid only in INTERACTIVE_MODE (returns -ENODEV if not).
//  NON-blocking write.  If there's room in the command buffer, the
//  user data is put there.  If not, we immediately return 0.  Note that
//  one can NEVER successfully write data whose length is greater than
//  the size of the card's command buffer!  (We return -EINVAL to so indicate.)
//
//  In order to avoid a user-program consuming CPU cycles by busily
//  re-issuing a write(2) syscall that return 0, we delay 2 jiffies
//  before we return 0.
int oms_fop_write( struct file* pfile,
                   const char* userbuf,
                   size_t usersiz, 
                   loff_t *ppos )
{
	int               minor;
	struct oms_card*  card;
	struct list_head* ptr;
	int               rv;
	char*             tmp;

	ASSERT_UNLOCKED();
	
	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;

	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}
	
	// ensure card is in interactive mode
	if ( card->comm_mode != INTERACTIVE_MODE )
		return -ENODEV;
        
	// make sure we have room for the command and the command exists...
	if ( usersiz >= INTERACTIVE_BUFFER_LEN ||
			 usersiz < 1 ||
			 usersiz > _oms_xmit_space_avail( card ) )
		return -EINVAL;

	// grab temporary buffer and copy data in before locking
	tmp = kmalloc( INTERACTIVE_BUFFER_LEN, GFP_KERNEL );
	if ( copy_from_user( tmp, userbuf, usersiz ) )
		return -EFAULT;

	// perform card-specific write operation
	_LOCK_(flags);
	rv = _oms_fop_write( card, tmp, usersiz );
	UN_LOCK_(flags);

	kfree( tmp );

	return rv;
}
//****************************************************************************


//****************************************************************************
// oms_fop_poll() --- support select(2)/poll(2) syscalls for read(2) ONLY.
//   (We don't support polling for write(2).  Instead, a write(2) will
//   fail, i.e., return 0, if there is insufficient room in card command
//   buffer.
unsigned int oms_fop_poll( struct file* pfile,
                           struct poll_table_struct* polltable )
{
	int               minor;
	struct oms_card*  card;
	struct list_head* ptr;
	unsigned int      rv = 0;

	ASSERT_UNLOCKED();

	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}
	
	// ensure card is in interactive mode
	if ( card->comm_mode != INTERACTIVE_MODE )
	{
		// hmmmm ... say device is readable, since read(2)
		//   can return -ENODEV and we can't
		return POLLIN | POLLRDNORM;
	}
	
	// tell the upper-layers of select(2) what waitq will be
	//   awakened when text data is available in the device.
	//   Our IRQ handler will issue a wake_up() on
	//   this waitq if card->comm_mode == INTERACTIVE_MODE.
	if ( card->response_ready == FALSE )
	{
		poll_wait( pfile, &card->poll_waitq, polltable );
	}
	
	_LOCK_(flags);
    // if there is data in the response buffer then
    //   return value indicating that device is now readable
	if ( _oms_recv_space_used( card ) )
		rv |= POLLIN | POLLRDNORM;
	UN_LOCK_(flags);

	return rv;
}
//****************************************************************************

	
//****************************************************************************
int oms_fop_ioctl( struct inode* pinode, 
                   struct file* pfile,
                   unsigned int cmd,
                   unsigned long userarg )
{
	int                   minor;
	struct oms_card*      card;
	struct list_head*     ptr;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
	void*                 arg = (void*)userarg;

	ASSERT_UNLOCKED();

	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;

	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}

	// execute requested ioctl function
	switch (cmd) 
	{
		case IOCTL_OMS_GET_VERSION:
		{
			struct oms_ioctl sioctl;
			int              len = strlen(version_string) + 1;
			
			if ( copy_from_user( &sioctl, (void*)arg, sizeof(sioctl) ) )
				return -EFAULT;
			if ( sioctl.response_len < len )
				return -EINVAL;
			if ( copy_to_user( sioctl.response, version_string, len ) )
				return -EFAULT;
			return 0;  // no tasklet start
		}
		
		case IOCTL_OMS_COMMAND:
		{
			// Send a command to the oms card. Invalid in interactive mode.
			if ( card->comm_mode == INTERACTIVE_MODE )
				return -ENODEV;

			pioctl = oms_new_ioctl( card );
			if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
			{
				return rv;  // ioctl already released
			}

			pioctl->cmd = cmd;
			pioctl->response_expected = FALSE;
			pioctl->comm_mode = STRING_MODE;
			
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  enqueueing COMMAND\n", me );
				
			_LOCK_(flags);
			tasklet_schedule( &oms_tasklet );
			rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
			UN_LOCK_(flags);
			
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  dequeueing COMMAND\n", me );

			oms_release_ioctl( pioctl );
			return (rv);
		}

		case IOCTL_OMS_COMMAND_NONALPHA:
		{
			// Send a binary string to the oms card. Invalid in interactive mode
			if ( card->comm_mode == INTERACTIVE_MODE )
				return -ENODEV;

			pioctl = oms_new_ioctl( card );
			if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
			{
				return rv;  // ios already released
			}

			pioctl->cmd = cmd;
			pioctl->response_expected = FALSE;
			pioctl->comm_mode = BINARY_MODE;
			
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  enqueueing COMMAND_NONALPHA\n", me );

			_LOCK_(flags);
			tasklet_schedule( &oms_tasklet );
			rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
			UN_LOCK_(flags);

			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  dequeueing COMMAND_NONALPHA\n", me );

			oms_release_ioctl( pioctl );
			return (rv);
		}

		case IOCTL_OMS_QUERY:
		{
			// Send a command to the oms card and await a response
			// Invalid in interactive mode
			if ( card->comm_mode == INTERACTIVE_MODE )
				return -ENODEV;
			
			pioctl = oms_new_ioctl( card );
			if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 1, arg )) != 0 )
			{
				return rv;  // ios already released
			}

			// incorporate operation-specific info into ioctl
			pioctl->cmd = cmd;
			pioctl->response_expected = TRUE;
			pioctl->comm_mode = STRING_MODE;
			
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  enqueueing QUERY\n", me );
			
			_LOCK_(flags);
			tasklet_schedule( &oms_tasklet );
			rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
			UN_LOCK_(flags);

			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  dequeueing QUERY\n", me );

			if ( !rv ) 
			{
				if ( DEBUGLEVEL > 2 )
					printk( KERN_INFO "%s:  QUERY response \"%s\"\n", me, pioctl->buffers.response );
	
				rv = oms_copy_out_ioctl_response( pioctl, arg );
			}
			
			oms_release_ioctl( pioctl );
			return (rv);
		}
		
		case IOCTL_OMS_QUERY_NONALPHA:
		{
			// Send a binary string to the oms card and await a response.
			// Invalid in interactive mode
			if ( card->comm_mode == INTERACTIVE_MODE )
				return -ENODEV;
			
			pioctl = oms_new_ioctl( card );
			if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 1, arg )) != 0 )
			{
				return rv;  // ios already released
			}

			// incorporate operation-specific info into ios
			pioctl->cmd = cmd;
			pioctl->response_expected = TRUE;
			pioctl->comm_mode = BINARY_MODE;
			
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  enqueueing QUERY_NONALPHA\n", me );

			_LOCK_(flags);
			tasklet_schedule( &oms_tasklet );
			rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
			UN_LOCK_(flags);

			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  dequeueing QUERY_NONALPHA\n", me );

			if ( !rv ) 
			{
				if ( DEBUGLEVEL > 2 )
					printk( KERN_INFO "%s:  QUERY_NONALPHA response \"%s\"\n", me, pioctl->buffers.response );
	
				rv = oms_copy_out_ioctl_response( pioctl, arg );
			}
			
			oms_release_ioctl( pioctl );
			return (rv);
		}
		
		case IOCTL_OMS_READ_SHARED_MEM:
			return _oms_ioctl_read_shared_mem( card, (void*)arg );
		
		case IOCTL_OMS_WRITE_SHARED_MEM:
			return _oms_ioctl_write_shared_mem( card, (void*)arg );
		
		case IOCTL_OMS_READ_STATUS_FLAGS:
			return _oms_ioctl_read_status_flags( card, arg );

		case IOCTL_OMS_CLEAR_STATUS_FLAGS:
			return _oms_ioctl_clear_status_flags( card, arg );
		
		case IOCTL_OMS_READ_DONE_FLAGS:
			return _oms_ioctl_read_done_flags( card, arg );

		case IOCTL_OMS_CLEAR_DONE_FLAGS:
			return _oms_ioctl_clear_done_flags( card, arg );
		
		case IOCTL_OMS_READ_PLIMIT_FLAGS:
			return _oms_ioctl_read_pos_limit_flags( card, arg );

		case IOCTL_OMS_CLEAR_PLIMIT_FLAGS:
			return _oms_ioctl_clear_pos_limit_flags( card, arg );
		
		case IOCTL_OMS_READ_NLIMIT_FLAGS:
			return _oms_ioctl_read_neg_limit_flags( card, arg );

		case IOCTL_OMS_CLEAR_NLIMIT_FLAGS:
			return _oms_ioctl_clear_neg_limit_flags( card, arg );
		
		case IOCTL_OMS_READ_SLIP_FLAGS:
			return _oms_ioctl_read_slip_flags( card, arg );

		case IOCTL_OMS_CLEAR_SLIP_FLAGS:
			return _oms_ioctl_clear_slip_flags( card, arg );
		
		case IOCTL_OMS_READ_AXIS_POSITIONS:
			return _oms_ioctl_read_axis_positions( card, (void*)arg );

		case IOCTL_OMS_CMD_BUFF_FREE:
		{
			// perform the arithmetic immediately, without queueing
			//  an ios through the system.
			pioctl = oms_new_ioctl( card );
			if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 4, arg )) != 0 )
				return rv;   // ios already released
			// the response area must be exactly 4 bytes long
			if ( pioctl->buffers.response_len != sizeof(u32) ) 
			{
				oms_release_ioctl( pioctl );
				return -EINVAL;
			}
			
			_LOCK_(flags);
			*((u32*)(pioctl->buffers.response)) = (u32)(_oms_xmit_space_avail( card ));
			UN_LOCK_(flags);

			pioctl->p_response = sizeof(u32);
			pioctl->response_expected = 1; // fake out our copy_out func
			rv = oms_copy_out_ioctl_response( pioctl, arg );
			oms_release_ioctl( pioctl );
			return rv;
		}
		
		case IOCTL_OMS_ENTER_INTERACTIVE_MODE:
		{
			rv = 0;
			_LOCK_(flags);
			if ( card->open_count != 1 ) 
			{
				rv = -EBUSY;
			}
			else 
			{
				card->comm_mode = INTERACTIVE_MODE;
			}
			UN_LOCK_(flags);
			printk( KERN_INFO "%s:  Entered interactive mode (single open only)\n", me );
			return rv;
		}
		
		case IOCTL_OMS_EXIT_INTERACTIVE_MODE:
		{
			rv = 0;
			_LOCK_(flags);
			if ( card->open_count != 1 ) 
			{
				rv = -EBUSY;
			}
			else 
			{
				card->comm_mode = STRING_MODE;
			}
			UN_LOCK_(flags);
			printk( KERN_INFO "%s:  Exited interactive mode\n", me );
			return rv;
		}
		
		case IOCTL_OMS_REINIT:
		{
			// should only be called after a firmware upgrade
			return _oms_ioctl_reinit( card, arg );
		}                       

		default:
			return _oms_fop_ioctl( pinode, pfile, cmd, userarg );
	}
	return 0;
}
//****************************************************************************


//****************************************************************************
// oms_tasklet_ioctl() --- function run in a TASKLET (aka DPC).
//   For all cards in the list, call oms_start_ioctl().
void oms_tasklet_ioctl( unsigned long ignored )
{
	struct list_head* pos;
	struct oms_card*  card;
	
	ASSERT_UNLOCKED();
	
	_LOCK_(flags);
	list_for_each( pos, &oms_cardlist ) 
	{
		card = list_entry( pos, struct oms_card, listhead );
		oms_start_ioctl( card );
	}
	UN_LOCK_(flags);
}
//****************************************************************************


//****************************************************************************
// oms_start_ioctl() -- examine the top ioctl structure in a card's
//     queue, and test whether the card has enough space to accept
//     the command.  If so, dequeue the ioctl structure and transfer
//     the command data to the card.  If the struct is marked as "done
//     at xfer", then complete the operation:  (1) if the structure is
//     marked having a task asleep, wake him up, else (2) simply
//     free the resources associated with this ioctl.
//
//     If there is insufficient room in the card for the command string,
//     set ios->userstruct.command_len to zero and arrange for the
//     sleeping task to transfer this negative result to the user program.
//
//  biglock is LOCKED
void oms_start_ioctl( struct oms_card* card )
{
	struct omsdrv_ioctl*   pioctl;
	struct list_head*      pos;
	
	ASSERT_LOCKED();

	// if the card is flagged 'blocking', just return
	if ( card->blocking == TRUE ) 
	{
		if ( DEBUGLEVEL > 0 )
			printk( KERN_INFO "%s:  card blocking\n", me );
		return;
	}
	
	if ( !list_empty( &card->ioctl_list ) )
	{
		pos = card->ioctl_list.next;
		pioctl = list_entry( pos, struct omsdrv_ioctl, listhead );
		if ( pioctl->comm_mode != INTERACTIVE_MODE ) 
		{
			// processing a string/binary, block any others that come along
			card->blocking = TRUE;

			if ( DEBUGLEVEL > 0 )
				printk( KERN_INFO "%s:  card set to blocking\n", me );

			// Do any card-specific operations required
			_oms_start_ioctl( card );
		}
	}
}
//****************************************************************************


//****************************************************************************
// oms_ioctl_complete
// Returns TRUE iff command has been completely transmitted and any expected
// reply has been received
inline BOOL oms_ioctl_complete( struct omsdrv_ioctl* pioctl )
{
	if ( pioctl->buffers.command_len &&
			 ( pioctl->p_command < pioctl->buffers.command_len ) )
		return FALSE;

	if ( pioctl->response_expected )
		return pioctl->card->response_ready;

	return TRUE;
}
//****************************************************************************


//****************************************************************************
// oms_enqueue_ioctl_and_sleep() --- called from the oms_fop_ioctl() function
//    (a file_operations routine running the user thread).  Enqueue an
//    ioctl op and sleep, with timeout, until awakened.  Return a
//    negative errno on error, or 0.
//
//  Called with spinlock 'biglock' LOCKED, with flags in reference var pflags.
//    (because we can block, we need to release and re-acquire the lock)
int oms_enqueue_ioctl_and_sleep( struct omsdrv_ioctl* pioctl,
                                 unsigned long* pflags )
{
	struct oms_card*  card = pioctl->card;
	int               rv;

	ASSERT_LOCKED();

	if (card == NULL) 
	{
		printk(KERN_ERR "%s:  NULL card pointer in ios\n", me);
		return -EIO;  // general purpose error code
	}

	// enqueue the ioctl
	INIT_LIST_HEAD( &pioctl->listhead );
	init_waitqueue_head( &pioctl->waitq );
	pioctl->task_asleep = 1;
	pioctl->card->response_ready = 0;
	list_add_tail( &pioctl->listhead, &card->ioctl_list );
	UN_LOCK_(*pflags);
	
	// sleep until awakened or timed out
	tasklet_schedule( &oms_tasklet );
	rv = sleep_on_timeout( &pioctl->waitq, IO_TIMEOUT * HZ );
	
	// awake, clean up
	_LOCK_(*pflags);
	if ( DEBUGLEVEL > 0 )
		printk( KERN_INFO "%s:  card set to non-blocking\n", me );
	card->blocking = 0;
	pioctl->task_asleep = 0;

	if ( oms_ioctl_complete( pioctl ) )
		return 0;

	if (rv == 0)
		return -ETIME;

	return -EIO;   // catchall.
}
//****************************************************************************


//****************************************************************************
// oms_copy_out_ioctl_response() -- copy the user's 'struct oms_ioctl'
//   structure, and if there's a response, copy that out too.
//   Return 0 on success or -EFAULT.
//
//   (we must NOT be locked, as one should never access user space under
//    a spinlock --- there might be a page fault during transfer)
int oms_copy_out_ioctl_response(	struct omsdrv_ioctl* pioctl,
                                  void* usr_arg )
{
	struct oms_ioctl  tmp_ioctl;
	int               extra = 0;
	
	ASSERT_UNLOCKED();

	// only do the work if the ioctl asked for a response
	if ( pioctl->response_expected && pioctl->p_response) 
	{
		// get the user struct first	
		copy_from_user( &tmp_ioctl, usr_arg, sizeof(struct oms_ioctl) );

		// if not interactive and a response exists, make sure it's nul terminated
		if ( pioctl->comm_mode != INTERACTIVE_MODE &&
				 pioctl->p_response < pioctl->buffers.response_len ) 
		{
			pioctl->buffers.response[pioctl->p_response] = 0;  // nul byte
			extra = 1;
		}

		// copy out the response itself (tmp_ioctl.response is a pointer to user space)
		if ( copy_to_user( tmp_ioctl.response,
                       pioctl->buffers.response,
                       pioctl->buffers.response_len + extra ) ) 
		{
			return -EFAULT;
		}

		// on input, the user's response_len is the max size of his buffer.
		// on output, response_len is the actual byte count copied out (less the
		//  nul byte if in STRING_MODE).
		tmp_ioctl.response_len = pioctl->buffers.response_len;

		// now copy the whole struct just to get the size copied...
		if ( copy_to_user( (void*)usr_arg,
	                     &tmp_ioctl,
	                     sizeof(struct oms_ioctl) ) ) 
		{
			return -EFAULT;
		}
	}
	
	return 0;
}
//****************************************************************************


