/* oms-core.h --- non-user-space header file
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 */

#ifndef __OMS_CORE_H_
#define __OMS_CORE_H_

#include <linux/ioctl.h>
#include "oms.h"		// user-space header

// structure used for queueing ioctl operations for sequential service
struct omsdrv_ioctl {
	unsigned long				cmd;					// copy of ioctl arg 2
  struct oms_card*		card;					// pointer to card this is for
        
	struct oms_ioctl		buffers;    	// copied in from user-space
	int									p_command;		// index in buffers.command of next char to xmit
	unsigned char     	response_expected;  // NO_RESPONSE or TEXT_RESPONSE
	int									p_response;		// index in buffers.response of next char to recv
		
	struct list_head		listhead;			// list per card
	wait_queue_head_t 	waitq;				// process sleeps here

	unsigned char     	comm_mode;		// STRING... or INTERACTIVE.. or BIN..
	unsigned char     	task_asleep;	// boolean:  TRUE if task sleeping
	int									lf_count;			// number of LFs received thus far
};

#define IOCTL_NOT_SUPPORTED			-EINVAL

#define	INTERACTIVE_BUFFER_LEN	1024
#define IO_TIMEOUT     					10     // number of seconds

// per-device instance of a card
// 
//  Minor number and uniqueness conventions:
//       1.  each card must have a unique DIP switch setting.
//       2.  the minor number of the device is the DIP switch setting.
//     (explanatory note:  the number generated by the DIP switches
//      is added to PCI_DEVICE_ID_OMS_BASE and appears in the
//      PCI config header field "subsystem device id".)
struct oms_card {
	struct list_head listhead;			// for driver-wide list of cards
	struct list_head ioctl_list;		// list of IOCTLs for this card
        
	struct pci_dev*  pdev;					// system PCI device structure
	int							 minor;					// minor number for this card
	int							 irq;
				
	int              blocking;			// whether handling a blocking IOCTL
	int							 open_count;		// number of open handles to this card

	#define STRING_MODE				0
	#define INTERACTIVE_MODE	1
	#define BINARY_MODE				2
	short          comm_mode;				// current communication mode

	// INTERACTIVE_MODE buffers
	char					 xmit_buffer[INTERACTIVE_BUFFER_LEN];
	int						 xmit_ins_ptr;
	int						 xmit_rem_ptr;
	char					 recv_buffer[INTERACTIVE_BUFFER_LEN];
	int						 recv_ins_ptr;
	int						 recv_rem_ptr;
	int						 lf_count;
	BOOL					 response_ready;
		
	// we need a wait queue to support select(2)/poll(2) syscall
	wait_queue_head_t poll_waitq;

	void*						hardware;				// pointer to card-specific data
};




//***************************************************************
//  A single spinlock, and conditional code for spinlock debugging
static spinlock_t  oms_biglock;

							  

// on a uniprocessor, we can make an ASSERT_UNLOCKED() test.  Since
//   RedHat kernels are configured with CONFIG_SMP defined, we can't use
//   <linux/config.h> to tell if we're running on a uniprocessor.  We
//   therefore make the assertion a MODULE_PARM
#define OMS_DEBUG_SPINLOCKS 1
#if OMS_DEBUG_SPINLOCKS
	#define _LOCK_(f) \
  	({spin_lock_irqsave(&oms_biglock, (f)); \
    if (atomic_read(&oms_islocked) != 0) \
    	printk(KERN_ERR "%s:  LOCK() bad %s:%d\n", me, __FILE__, __LINE__); \
		atomic_set(&oms_islocked, 1);})
	#define UN_LOCK_(f) \
  	({if (atomic_read(&oms_islocked) != 1) \
    	printk(KERN_ERR "%s:  UNLOCK() bad %s:%d\n", me,__FILE__,__LINE__); \
      atomic_set(&oms_islocked, 0); \
      spin_unlock_irqrestore(&oms_biglock, (f));})
	#define ASSERT_LOCKED() \
  	({if (atomic_read(&oms_islocked) != 1) \
    		printk(KERN_ERR "%s:  ASSERT_LOCKED bad %s:%d\n", me,__FILE__, __LINE__);})
	#define ASSERT_UNLOCKED() \
  	({if (ONE_CPU && atomic_read(&oms_islocked) != 0) \
			printk(KERN_ERR "%s:  ASSERT_UNLOCKED bad %s:%d\n", me,__FILE__, __LINE__);})
#else // if NOT OMS_DEBUG_SPINLOCKS --- vanilla lock/unlock without tests
	#define _LOCK_(f)     spin_lock_irqsave(&oms_biglock, (f))
	#define UN_LOCK_(f)   spin_unlock_irqrestore(&oms_biglock, (f))
	#define ASSERT_LOCKED()   /* empty */
	#define ASSERT_UNLOCKED() /* empty */
#endif // OMS_DEBUG_SPINLOCKS


//  ********* function prototypes **************
int    oms_install_module( void );
int    oms_install_one( struct pci_dev *pdev,
                        const struct pci_device_id *ent );
int    oms_attach_irq( int irq );
void   oms_remove_module( void );
void   oms_remove_one( struct pci_dev *pdev );
void   oms_release_resources( struct oms_card* card );
void   oms_create_nodes(void);
void   oms_tasklet_ioctl( unsigned long ignored );
void   oms_start_ioctl( struct oms_card* card );
struct omsdrv_ioctl* oms_new_ioctl( struct oms_card* card );
void   oms_release_ioctl( struct omsdrv_ioctl* );
int    oms_fetch_user_ioctl_struct( struct omsdrv_ioctl* pioctl,
                                    int command_parm,
                                    int response_parm,
                                    void* user_arg );
int    oms_enqueue_ioctl_and_sleep( struct omsdrv_ioctl* pioctl,
                                    unsigned long* pflags );
int    oms_copy_out_ioctl_response( struct omsdrv_ioctl* pioctl,
                                    void* user_arg );

// char device file operations
int          oms_fop_open( struct inode*, struct file* );
int          oms_fop_close( struct inode*, struct file* );
int          oms_fop_ioctl( struct inode*, struct file*, unsigned int, unsigned long );
int          oms_fop_read( struct file*, char*, size_t, loff_t* );
int          oms_fop_write( struct file*, const char*, size_t, loff_t* );
unsigned int oms_fop_poll( struct file*, struct poll_table_struct* );

void   oms_sleep( int jifs );


//  **** function prototypes which MUST be supported by card-specific code ****
irqreturn_t _oms_irq_handler( int irq, void *dev, struct pt_regs *regs );
int  _oms_init_memory( struct pci_dev* pdev,
                       struct oms_card* card );
int  _oms_attach_irq( struct pci_dev* pdev,
                      struct oms_card* card );
void _oms_detach_irq( struct oms_card* card );
void _oms_release_resources( struct oms_card* card );
int  _oms_fop_read( struct oms_card* card,
                    char buffer[],
                    int maxsize );
int  _oms_fop_write( struct oms_card* card,
                     char buffer[],
                     int maxsize );
int  _oms_fop_ioctl( struct inode*, struct file*, unsigned int, unsigned long );

int  _oms_xmit_space_avail( struct oms_card* card );
int  _oms_recv_space_used( struct oms_card* card );
void _oms_start_ioctl( struct oms_card* card );
int  _oms_ioctl_read_status_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_clear_status_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_read_done_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_clear_done_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_read_pos_limit_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_clear_pos_limit_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_read_neg_limit_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_clear_neg_limit_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_read_slip_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_clear_slip_flags( struct oms_card* card, void* arg );
int  _oms_ioctl_read_shared_mem( struct oms_card* card, void* arg );
int  _oms_ioctl_write_shared_mem( struct oms_card* card, void* arg );
int  _oms_ioctl_read_axis_positions( struct oms_card* card, void* arg );
int  _oms_ioctl_reinit( struct oms_card* card, void* arg );


#endif // __OMS_CORE_H_
