/*
 *  oms-pcixdrv.c --- Linux device driver for Oregon Micro Systems
 *    PCIx motion controller, hardware-specific code.
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 */

char version_string[] = "Oregon Micro Systems PCIx Linux Driver v2.0";
char* me = "PCIx";

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/config.h>
#include <linux/version.h>

#ifdef CONFIG_MODVERSIONS
#ifndef MODVERSIONS
#define MODVERSIONS
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,40)
#include <linux/modversions.h>
#endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/kmod.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <linux/byteorder/generic.h>

#include "oms-core.h"      // driver-private header
#include "oms-pcixdrv.h"   // driver-private card-specific header

MODULE_DESCRIPTION("Driver for OMS PCIx motion control");
char* node_device = "pcix";

//****************************************************************************
// PCI BUS VENDOR AND DEVICE IDS FOR BUS INTERFACE AND OMS DEVICES

// The bus interface vendor/device IDs will be in system vendor ID
//  registers of PCI configuration space
#define PCI_VENDOR_ID_BUS	    	0x10B5
#define PCI_DEVICE_ID_BUS				0x9050
// The OMS vendor/device IDs will be in the subsystem vendor ID
//   registers of PCI configuration space
#define PCI_VENDOR_ID_OMS				0x160c
#define PCI_DEVICE_NAME					"pcix"

int PCI_DEVICE_ID_OMS_BASE = 0x0010;


// pci_device_id --- specify to PCI driver the set of vendor/device IDs
//     that we know about.  The PCI driver will call our "probe" routine
//     for each device matching one of our descriptions.
struct pci_device_id oms_pci_tbl[] __devinitdata = 
{
	{ PCI_VENDOR_ID_BUS,
	  PCI_DEVICE_ID_BUS,
	  PCI_VENDOR_ID_OMS,
	  PCI_ANY_ID    // we will accept anything >= PCI_OMS_DEVICE_ID_BASE
	},
	{ 0 }
};


// pci_driver structure --- register with the PCI driver.  It will call
//   our "probe" routine for each PCI device matching our table.
struct pci_driver oms_driver = 
{
	.name     = PCI_DEVICE_NAME,
	.id_table = oms_pci_tbl,
	.probe    = oms_install_one,
	.remove   = __devexit_p(oms_remove_one),
};



//****************************************************************************
// externs from oms-core.c
extern struct list_head oms_cardlist;
extern unsigned long flags;
// Module parameters
extern int DEBUGLEVEL;
extern int MAJORNUMBER;
extern int ONE_CPU;
extern int MAX_CARDS;
extern char* NODE_CREATION_SCRIPT;
#if OMS_DEBUG_SPINLOCKS
	extern atomic_t   oms_islocked;
#endif


//****************************************************************************
// Private functions
BOOL _oms_irq_handle_tbe( struct oms_card* card );
void _oms_irq_handle_ibf( struct oms_card* card );


//****************************************************************************
// _oms_init_memory ---- called by oms_install_one to probe for
//   card-specific memory regions and allocate them
//
//  We don't know the index numbers of the 8 byte and 8k BARs we need.
int _oms_init_memory( struct pci_dev* pdev,
										  struct oms_card* card )
{
	int             	i;
	int             	bars_found = 0;
	unsigned long   	start, len, siz;
	int								bars_128 = 0;
	char							char1, char2;
	int 							char_matches = 0;

	card->hardware = (void*)kmalloc( sizeof(struct hw_struct), GFP_ATOMIC );
	memset(card->hardware, 0, sizeof(struct hw_struct));
	
	// scan the BARs looking for MMIO regions of 128 bytes and 8k
	for (i = 0; bars_found < 2 && i < DEVICE_COUNT_RESOURCE; i++) 
	{
		start = pci_resource_start(pdev, i);
		len   = pci_resource_len(pdev, i);

		if (start == 0 || len == 0)
			continue;
		if ((pci_resource_flags(pdev, i) & IORESOURCE_MEM) == 0)
			continue;

		siz = 128;  // look for 128 bytes of register space
		// There are 2 128 byte BARs for the PCIx. The first is the PCI
		// configuration space, the second is the remapped register region.
		if (len == siz )
		{
			if ( ++bars_128 < 2 )
				continue;

			if (!request_mem_region(start, siz, me)) 
			{
				printk( KERN_ERR "%s:  mem region collision at %p\n", me, (void*) start );
				return 1;
			} 
			// we store the base of the mapped registers
			//  as an unsigned long, as we never dereference
			//  it, but need the value for address arithmetic
			//  to derive the pointers we actually use.
			card_hw->regbase_addr = (unsigned long)ioremap(start, siz);

			if (!card_hw->regbase_addr) 
			{
				printk( KERN_ERR "%s:  ioremap failed\n", me );
				return 1;
			}
			card_hw->regbase_physaddr = start;
			// set pointers to registers within the 128 byte space
			card_hw->reg_control = (u8 *) (card_hw->regbase_addr + OMS_REG_CONTROL);
			card_hw->reg_status  = (u8 *) (card_hw->regbase_addr + OMS_REG_STATUS);
			card_hw->reg_done    = (u8 *) (card_hw->regbase_addr + OMS_REG_DONE);
			card_hw->reg_data    = (u8 *) (card_hw->regbase_addr + OMS_REG_DATA);
			bars_found++;
			// continue 'for' loop
			continue;
		}

		siz = 8 * 1024;   // look for shared memory
		if (len == siz) 
		{
			if (!request_mem_region(start, siz, me)) 
			{
				printk( KERN_ERR "%s:  mem region collision at %p\n", me, (void*) start );
				return 1;
			} 
			card_hw->sharedmem = ioremap(start, siz);
			if (!card_hw->sharedmem) 
			{
				printk(KERN_ERR "%s:  ioremap failed\n", me);
				return 1;
			}
			card_hw->sharedmem_physaddr = start;
			bars_found++;
			// continue 'for' loop
			continue;
		}
	}
	if (bars_found < 2) 
	{
		printk( KERN_ERR "%s:  failed to find all needed BARs\n", me );
		return 1;
	}

	// Set up pointers to commonly used shared memory structures
	card_hw->cardnum        = (u16*) &card_hw->sharedmem[OMS_MEM_CARDNUM];
	card_hw->mailbox        = (u16*) &card_hw->sharedmem[OMS_MEM_MAILBOX];
	card_hw->vec_opcode     = (u16*) &card_hw->sharedmem[OMS_MEM_VEC_OPCODE];
	card_hw->vec_insert     = (u16*) &card_hw->sharedmem[OMS_MEM_VEC_INSERT];
	card_hw->vec_process    = (u16*) &card_hw->sharedmem[OMS_MEM_VEC_PROCESS];

	// make sure the card's data buffer is empty
	char1 = *card_hw->reg_data;
	char_matches = 0;
	while ( char_matches < 10 )
	{
		char2 = *card_hw->reg_data;
		if ( char1 == char2 )
			char_matches++;
		else
			char_matches = 0;
		char1 = char2;
	}
	
	return 0;
}
//****************************************************************************


//****************************************************************************
// _oms_attach_irq()
//
//   Note:  spinlock oms_biglock is LOCKED on entry.
int _oms_attach_irq( struct pci_dev* pdev,
										 struct oms_card* card )
{
	int    rv;
	struct list_head* ptr;

	ASSERT_LOCKED();
	
	// see if we've already requested this IRQ
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		struct oms_card* a_card = list_entry( ptr, struct oms_card, listhead );
		if ( a_card->irq == pdev->irq )
		{
			return 0;
		}
	}

	if ( (rv = request_irq(pdev->irq, &_oms_irq_handler, SA_SHIRQ, me, card)) != 0) 
	{
		printk(KERN_ERR "%s:  request_irq failed on irq %d:  %d\n", me, pdev->irq, rv);
		return -1;
	}

	// now we can set the irq number in the card struct
	card->irq = pdev->irq;

	// initialize card's interrupt enable register
	*card_hw->reg_control = INTERRUPT_MASK;

	return 0;
}
//****************************************************************************


//****************************************************************************
// _oms_detach_irq()
//
//   Note:  spinlock oms_biglock is LOCKED on entry.
void _oms_detach_irq( struct oms_card* card )
{
	ASSERT_LOCKED();
	
	// free the irq
	(void)free_irq( card->irq, card );
}
//****************************************************************************


//****************************************************************************
// _oms_release_resources() --- release all resources used by one card
//
//  Note:  spinlock oms_biglock is LOCKED on entry.
void _oms_release_resources( struct oms_card* card )
{
	ASSERT_LOCKED();

	if ( card->hardware )
	{
		if ( card_hw->sharedmem ) 
		{
			iounmap( card_hw->sharedmem );
		}
		if ( card_hw->regbase_addr ) 
		{
			iounmap( (void*)card_hw->regbase_addr );
		}
		if ( card_hw->sharedmem_physaddr ) 
		{
			release_mem_region( card_hw->sharedmem_physaddr, (8 * 1024) );
		}
		if ( card_hw->regbase_physaddr ) 
		{
			release_mem_region( card_hw->regbase_physaddr, 128 );
		}
		kfree( card->hardware );
	}
	if ( card->irq >= 0 )
		_oms_detach_irq( card );
}
//****************************************************************************


//****************************************************************************
// _oms_fop_read() --- valid ONLY if card->comm_mode is INTERACTIVE_MODE.
//
// NON-blocking read.  If there's nothing to read, returns 0 immediately.
//     Use select(2) or poll(2) to avoid busy looping.  Each time we're
//     called we grab ALL data from the card into a temporary buffer, and
//     copy out what will fit in the user's buffer.  Any residue in
//     temporary buffer is silently discarded.  This is semi-okay, as
//     INTERACTIVE_MODE is intended only for use by OMS user programs,
//     which can be developed to allocate a large enough buffer.
int  _oms_fop_read( struct oms_card* card,
                    char buffer[],
                    int maxsize )
{
	int    rv = 0;
	
	ASSERT_LOCKED();
	
	if ( card->response_ready )
	{
		while ( rv <= maxsize && _oms_recv_space_used( card ) < INTERACTIVE_BUFFER_LEN )
		{
			buffer[rv++] = card->recv_buffer[card->recv_rem_ptr];
			card->recv_buffer[card->recv_rem_ptr] = 0;
			if ( ++card->recv_rem_ptr > INTERACTIVE_BUFFER_LEN )
				card->recv_rem_ptr = 0;
		}
		card->response_ready = FALSE;
	}
	return rv;
}
//****************************************************************************


//****************************************************************************
// _oms_fop_write --- valid only in INTERACTIVE_MODE (returns -ENODEV if not).
//  NON-blocking write.  If there's room in the command buffer, the
//  user data is put there.  If not, we immediately return 0.  Note that
//  one can NEVER successfully write data whose length is greater than
//  the size of the card's command buffer!  (We return -EINVAL to so indicate.)
//
//  In order to avoid a user-program consuming CPU cycles by busily
//  re-issuing a write(2) syscall that return 0, we delay 2 jiffies
//  before we return 0.
int  _oms_fop_write( struct oms_card* card,
                     char buffer[],
                     int maxsize )
{
	int     rv = 0;

	ASSERT_LOCKED();
	
	// Stuff the command in the outgoing buffer
	while ( rv < maxsize && buffer[rv] != 0 && (_oms_xmit_space_avail( card ) >= (maxsize - rv)) )
	{
		if ( buffer[rv] >= 0x20 || buffer[rv] == CR )
			card->xmit_buffer[card->xmit_ins_ptr] = buffer[rv];
		rv++;
		if ( ++card->xmit_ins_ptr >= INTERACTIVE_BUFFER_LEN )
			card->xmit_ins_ptr = 0;
	}

	if ( rv > 0 )
	{
		// enable TBE interrupts
		*card_hw->reg_control = (*card_hw->reg_control & ~TBE_BIT);
		*card_hw->reg_control = (*card_hw->reg_control | TBE_BIT);
	}
	else
	{
		// delay here so user doesn't have to
		oms_sleep(2);
	}
	return rv;
}
//****************************************************************************

	
//****************************************************************************
int  _oms_xmit_space_avail( struct oms_card* card )
{
	if ( card->xmit_ins_ptr >= card->xmit_rem_ptr )
		return INTERACTIVE_BUFFER_LEN - (card->xmit_ins_ptr - card->xmit_rem_ptr);
	return card->xmit_rem_ptr - card->xmit_ins_ptr - 1;
}
//****************************************************************************


//****************************************************************************
int  _oms_recv_space_used( struct oms_card* card )
{
	if ( card->recv_ins_ptr >= card->recv_rem_ptr )
		return INTERACTIVE_BUFFER_LEN - (card->recv_ins_ptr - card->recv_rem_ptr);
	return card->recv_rem_ptr - card->recv_ins_ptr - 1;
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handler()
irqreturn_t _oms_irq_handler( int irq, void* dev, struct pt_regs* regs )
{
	irqreturn_t        retval;
	struct list_head*  pos;
	struct oms_card*   card;
	u8					       enabled_interrupts;
	u8                 interrupt_mask;
	u8					       curr_status;

	ASSERT_UNLOCKED();
	_LOCK_(flags);

	retval = IRQ_NONE;

	for ( pos = oms_cardlist.next; pos != &oms_cardlist; pos = pos->next )
	{
		card = list_entry( pos, struct oms_card, listhead );
                
		// Get the status of the enabled interrupts
		enabled_interrupts = *card_hw->reg_control;
		// Hold off any additional controller interrupt requests
		interrupt_mask = enabled_interrupts & 0x7f;
		*card_hw->reg_control = interrupt_mask;
   
		// Get the status of the enabled interrupts & the error flags
		curr_status = (*card_hw->reg_status) & interrupt_mask;
		if ( curr_status == 0 ) 
		{
			// nothing to do on this card.  carry on
			continue;
		}

		//If any error status flags are set
		if( curr_status & 0x0f ) 
		{
			//Preserve the error flags that were set
			//Note it's up to the application to explicitly clear the stored error
			// flags, via the IOCTL_OMS_CLR_STATUS IO control function, once they
			// have been processed. 
			card_hw->status_flags |= (curr_status & 0x0f);   
			retval = IRQ_HANDLED;
		}

		// If the DONE interrupt flag is set
		if( (curr_status & DONE_BIT) == DONE_BIT )
		{
			//OR the done flags in to the controller's done flag byte.
			//Note it's up to the application to explicitly clear the stored done
			// flags, via the IOCTL_OMS_CLR_DONE IO control function, once they have
			// been processed. 
			card_hw->done_flags |= *card_hw->reg_done;
			retval = IRQ_HANDLED;
		}
  
		// If the CONTROLLER has a response character waiting to be read
		if( (curr_status & IBF_BIT) == IBF_BIT )
		{
			_oms_irq_handle_ibf( card );
			retval = IRQ_HANDLED;
		}

	    //If the controller is ready to accept another command string character
		if( (curr_status & TBE_BIT) == TBE_BIT )
		{
			if ( _oms_irq_handle_tbe( card ) == FALSE )
			{
				// no more chars to send, disable TBE interrupts
				enabled_interrupts &= ~TBE_BIT;
			}
			retval = IRQ_HANDLED;
		}
		
		//Re-enable controller interrupts
		*card_hw->reg_control = enabled_interrupts;
	}
	
	UN_LOCK_(flags);

	return retval;
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handle_tbe() --- called from _oms_irq_handler() when
//   the card indicates that it has space available.  Determine from
//   the state of the card and the ioctl op queue what to do.
//
//  returns true iff there are more characters to send
//
//  Note:  the 'biglock' spinlock is locked.
BOOL _oms_irq_handle_tbe( struct oms_card* card )
{
	struct list_head*     pos;
	struct omsdrv_ioctl*  pioctl = NULL;
	char                  c_send = 0;
	BOOL                  b_more_to_send = FALSE;
	
	ASSERT_LOCKED();

	// if we're in interactive mode, send chars from the card's buffer
	if ( card->comm_mode == INTERACTIVE_MODE )
	{
		c_send = card->xmit_buffer[card->xmit_rem_ptr];
		card->xmit_buffer[card->xmit_rem_ptr] = 0;
		if ( ++card->xmit_rem_ptr >= INTERACTIVE_BUFFER_LEN )
			card->xmit_rem_ptr = 0;
		if ( _oms_xmit_space_avail( card ) != INTERACTIVE_BUFFER_LEN )
			b_more_to_send = TRUE;
	}
	else
	{
		// Non-interactive mode uses the buffers in the current ioctl
		// find the ios associated with the waiting command
		if ( list_empty( &card->ioctl_list ) ) 
		{
			// no ioctls waiting, why were we flagged to send a command?
			printk(KERN_ERR "%s:  irq:  no ios for expected command\n", me);
			return FALSE;
		}
		pos = card->ioctl_list.next;
		pioctl = list_entry( pos, struct omsdrv_ioctl, listhead );
	
		if ( !pioctl->buffers.command || pioctl->p_command >= pioctl->buffers.command_len ) 
		{
			if ( DEBUGLEVEL > 2 )
				printk( KERN_INFO "%s:  TBE - no command (%x) or command already sent (%d,%d)\n", me, 
						(unsigned int)pioctl->buffers.command,
						pioctl->p_command,
						pioctl->buffers.command_len );
			// an ioctl is present but command has been sent
			if ( pioctl->response_expected == FALSE )
			{
				if ( DEBUGLEVEL > 2 )
					printk( KERN_INFO "%s:  TBE - no response expected, waking task\n", me );
				// done, kill off the ios
				if ( pioctl->task_asleep )
					wake_up( &pioctl->waitq );
				else
				{
					oms_release_ioctl( pioctl );
				}
			}
			return FALSE;
		}
		c_send = pioctl->buffers.command[pioctl->p_command++];
		// always claim there is more to send. Allow the logic above to
		// complete the transaction
		b_more_to_send = TRUE;
	}

	if ( DEBUGLEVEL > 0 )
	{
		if ( card->comm_mode != INTERACTIVE_MODE )
		{
			if ( pioctl->p_command < 3 || 
					 pioctl->p_command >= pioctl->buffers.command_len ||
					 pioctl->buffers.command_len < 20 )
				printk( KERN_INFO "%s:  TBE [%02X] (%d of %d)\n", 
						me, 
						c_send,
						pioctl->p_command,
						pioctl->buffers.command_len );
			else if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  TBE [%02X]\n", me, c_send );
		}
		else
		{
			printk( KERN_INFO "%s:  TBE [%02X]\n", me, c_send );
		}
	}
		
	// send the next character
	*card_hw->reg_data = c_send;
	return ( b_more_to_send );
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handle_ibf() --- called from _oms_irq_handler() when
//   the card indicates that there is a TEXT_RESPONSE.  Determine from
//   the state of the card and the ioctl op queue what to do with the
//   data.
//
//  Note:  the 'biglock' spinlock is locked.
void _oms_irq_handle_ibf( struct oms_card* card )
{
	struct list_head*     pos;
	struct omsdrv_ioctl*  pioctl = NULL;
	u8                    resp_char;
	
	ASSERT_LOCKED();

	// read the input buffer regardless of what we are expected to do with it	
	resp_char = *card_hw->reg_data;

	if ( DEBUGLEVEL > 0 )
		printk( KERN_INFO "%s:  IBF [%02X]\n", me, resp_char );

	if ( card->comm_mode == INTERACTIVE_MODE )
	{
		if ( resp_char == LF )
		{
			if ( ++card->lf_count > 1 )
			{
				card->lf_count = 0;
				card->response_ready = TRUE;
				wake_up( &card->poll_waitq );
			}
		}
		else if ( resp_char != CR )
		{
			card->recv_buffer[card->recv_ins_ptr] = resp_char;
			if ( ++card->recv_ins_ptr >= INTERACTIVE_BUFFER_LEN )
				card->recv_ins_ptr = 0;
		}
	}
	else
	{
		// find the ioctl associated with the last command sent
		if ( list_empty( &card->ioctl_list ) ) 
		{
			// no ioctls waiting
			return;
		}
		pos = card->ioctl_list.next;
		pioctl = list_entry( pos, struct omsdrv_ioctl, listhead );
		if ( pioctl->response_expected == FALSE ) 
		{
			// an ios is present but doesn't want the response
			if ( pioctl->buffers.command && 
					 pioctl->p_command < pioctl->buffers.command_len )
			{
				if ( DEBUGLEVEL > 2 )
					printk( KERN_INFO "%s: IBF char received [%x] but still transmitting, discard\n", me, resp_char );
				// still transmitting, just exit
				return;
			}
			
			printk(KERN_ERR "%s: irq:  unexpected response from card\n", me);
			if ( pioctl->task_asleep ) 
			{
				wake_up( &pioctl->waitq );
			}
			else 
			{
				oms_release_ioctl( pioctl );
			}
			return;
		}

		// ioctl found.  store the received character and wake up the ioctl if
		// the response is complete.
		if ( pioctl->comm_mode == STRING_MODE )
		{			
			switch( resp_char )
			{
				case CR:
					// ignore carriage returns
					break;
				
				case LF:
					// 2 LFs == end of transmission
					if ( pioctl->lf_count > 0 )
					{
						pioctl->lf_count = 0;
						pioctl->card->response_ready = 1;
						
						// let the task complete the ioctl operation
						if ( pioctl->task_asleep ) 
						{
							wake_up( &pioctl->waitq );
						} 
						else 
						{
							printk(KERN_ERR "%s:  irq:  task not asleep for response\n", me);
							oms_release_ioctl( pioctl );
						}
		
						// done.					
					}
					else
						pioctl->lf_count++;
					break;
					
				default:
					pioctl->buffers.response[pioctl->p_response++] = resp_char;
					break;
			}
		}
		else	//BINARY_MODE
		{
			if ( DEBUGLEVEL > 1 )
				printk( KERN_INFO "%s:  IBF response status (%d of %d)\n", me, pioctl->p_response, pioctl->buffers.response_len );
			pioctl->buffers.response[pioctl->p_response++] = resp_char;
			if ( pioctl->p_response >= pioctl->buffers.response_len )
			{
				if ( DEBUGLEVEL > 1 )
					printk( KERN_INFO "%s:  IBF response complete (%d of %d)\n", me, pioctl->p_response, pioctl->buffers.response_len );
				pioctl->card->response_ready = 1;
				
				// let the task complete the ioctl operation
				if ( pioctl->task_asleep ) 
				{
					wake_up( &pioctl->waitq );
				} 
				else 
				{
					printk(KERN_ERR "%s:  irq:  task not asleep for response\n", me);
					oms_release_ioctl( pioctl );
				}

				// done.					
			}
		}
	}
}
//****************************************************************************


//****************************************************************************
// Support card-specific IOCTL operations
//
int  _oms_fop_ioctl( struct inode* pinode, 
                     struct file* pfile,
                     unsigned int cmd,
                     unsigned long userarg )
{
/*
	int                   minor;
	struct oms_card*      card;
	struct list_head*     ptr;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
	void*                 arg = (void*)userarg;

	ASSERT_UNLOCKED();

	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}
*/

	// execute requested ioctl function
	switch (cmd) 
	{
		default:
			return -EINVAL;
	}
	return 0;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_read_status_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the status_flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->status_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->status_flags) );  // pre-zero

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->status_flags, sizeof(card_hw->status_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->status_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_status_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear status flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->status_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->status_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_done_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the done_flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->done_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->done_flags) );

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->done_flags, sizeof(card_hw->done_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->done_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_done_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear done flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->done_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->done_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_pos_limit_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_pos_limit_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_neg_limit_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_neg_limit_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_slip_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_slip_flags( struct oms_card* card, void* arg )
{
	return IOCTL_NOT_SUPPORTED;
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_shared_mem( struct oms_card* card, void* arg )
{
	u16*									src;
	u16*									dest;
	u16										count, word, i, offst;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	if ( card->comm_mode == INTERACTIVE_MODE )
		return -ENODEV;

	// read shared memory.  The starting offset is the
	//   single u16 of the command string, the number
	//   of u16's to read is given by the response_len.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 2, 2, arg )) != 0 )
		return rv;  // pioctl already released

	// verify that command_len == sizeof(u16), i.e., one word
	//   also verify a non-zero response length
	if ( pioctl->buffers.command_len != sizeof(u16) ||	
			 pioctl->buffers.response_len == 0) 
	{
		oms_release_ioctl( pioctl );
			return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
	if (rv) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (rv);
	}
	// copy data from the shared memory address to ios->kresponse
	offst = *((u16*)pioctl->buffers.command);
	src = (u16*)(card_hw->sharedmem + offst);
	dest = (u16*)(pioctl->buffers.response);
	count = pioctl->buffers.response_len / sizeof(u16);
	// make sure src addr is word-aligned
	if ( (unsigned long)src & 0x01 ) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-EINVAL);
	}
	for ( i = 0; i < count; i++, src++ ) 
	{
		// range check
		if ( (unsigned long)src >=
						(((unsigned long)(card_hw->sharedmem)) +
						(8 * 1024)) ) 
		{
			UN_LOCK_(flags);
			oms_release_ioctl( pioctl );
			return (-EINVAL);
		}
		word = be16_to_cpu(readl(src));
		*dest++ = word;
	}
	UN_LOCK_(flags);

	pioctl->p_response = pioctl->buffers.response_len;
	pioctl->response_expected = TRUE;  // fake out copy_out function
	rv = oms_copy_out_ioctl_response( pioctl, arg );
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_write_shared_mem( struct oms_card* card, void* arg )
{
	u16*									src;
	u16*									dest;
	u16										count, word, i;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
			
	// write words (u16's) to shared memory.  The first
	//  word of the command string is the offset into
	//  shared memory.  The second word is the word count,
	//  and the third (and successive) words are the data
	//  to write in big-endian format.
	//
	//  This one is easy to get wrong in your user program.
	//  Make sure that command_len is:
	//             (2 * sizeof(__u16)) + sizeof(data)
	//  And that the 2 words of info (offset and WORD count)
	//  are contiguous with the data.
	//
	//  As an aid, we return -ENXIO if the command_len
	//  doesn't jibe with the word-count.
	if ( card->comm_mode == INTERACTIVE_MODE )
		return -ENODEV;
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 2, 0, arg )) != 0 )
		return rv;  // pioctl already released
	if ( pioctl->buffers.command_len < (3 * sizeof(u16)) ||
				pioctl->buffers.response_len != 0) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );

	// see if there was a timeout error or we must handle a sig
	if (rv) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return rv;
	}
	// copy the data into the shared memory
	src = (u16*)pioctl->buffers.command;
	dest = (u16*)(card_hw->sharedmem + *src++);
	count = *src++;  // must be a WORD count!!

	// make sure command_len jibes with the count
	if ( pioctl->buffers.command_len != (count + 2) * sizeof(u16) ) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-ENXIO);
	}

	// make sure that dest addr is word-aligned
	if ((unsigned long) dest & 0x01) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-EINVAL);
	}

	for ( i = 0; i < count; i++, dest++ ) 
	{
		word = *src++;
		// range check
		if ((unsigned long) dest >=
				(((unsigned long)(card_hw->sharedmem)) +
				(8 * 1024))) 
		{
			UN_LOCK_(flags);
			oms_release_ioctl( pioctl );
			return (-EINVAL);
		}
		writel( cpu_to_be16(word), dest );
	}
	UN_LOCK_(flags);
	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_read_axis_positions( struct oms_card* card, void* arg )
{		
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
	__u32*                kdata;
	unsigned long         timeout;
			
	if ( card->comm_mode == INTERACTIVE_MODE )
		return -ENODEV;

	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 4, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != (8 * sizeof(u32)) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}

	// tickle the mailbox to get updated position info
	*((unsigned short*)(card_hw->sharedmem + OMS_MEM_MAILBOX)) = MAILBOX_BOTH;
	timeout = jiffies + 3 * HZ;
	while ( *((unsigned short*)(card_hw->sharedmem + OMS_MEM_MAILBOX)) == MAILBOX_BOTH &&
			jiffies < timeout ){};

	kdata = (__u32*)pioctl->buffers.response;
	kdata[0] = *((__u32*)(card_hw->sharedmem + OMS_MEM_MOTOR_X));
	kdata[1] = *((__u32*)(card_hw->sharedmem + OMS_MEM_MOTOR_Y));
	kdata[2] = *((__u32*)(card_hw->sharedmem + OMS_MEM_MOTOR_Z));
	kdata[3] = *((__u32*)(card_hw->sharedmem + OMS_MEM_MOTOR_T));
	kdata[4] = *((__u32*)(card_hw->sharedmem + OMS_MEM_ENCODER_X));
	kdata[5] = *((__u32*)(card_hw->sharedmem + OMS_MEM_ENCODER_Y));
	kdata[6] = *((__u32*)(card_hw->sharedmem + OMS_MEM_ENCODER_Z));
	kdata[7] = *((__u32*)(card_hw->sharedmem + OMS_MEM_ENCODER_T));
	pioctl->p_response = 8 * sizeof(__u16);

	pioctl->response_expected = TRUE; // fake out our copy_out func
	rv = oms_copy_out_ioctl_response( pioctl, arg );

	if (rv == 0)
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	oms_release_ioctl( pioctl );
	return rv;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_reinit( struct oms_card* card, void* arg )
{
	// Rewrite the interrupt control register
	_LOCK_(flags);
	*card_hw->reg_control = INTERRUPT_MASK;
	UN_LOCK_(flags);

	return 0;
}
//****************************************************************************


//****************************************************************************
void _oms_start_ioctl( struct oms_card* card )
{
	ASSERT_LOCKED();

	// enable TBE interrupts
	*card_hw->reg_control = (*card_hw->reg_control & ~TBE_BIT);
	*card_hw->reg_control = (*card_hw->reg_control | TBE_BIT);
}
//****************************************************************************


