/*
 *  oms-maxp-lib.h --- define the user interface to the Oregon Mirco Systems
 *     library of functions that provide access to the MAXp board under
 *     Linux.
 *
 *  Copyright (c) 2003  Oregon Micro Systems.  All Rights Reserved.
 */

#ifndef __OMS_MAXP_LIB_H_
#define __OMS_MAXP_LIB_H_

// some types
typedef void *OMSHANDLE;    // returned by GetOmsHandle() and used in
                         //  all subsequent calls to the device.

#ifndef BOOL
#define BOOL int
#endif
#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif

// return codes for the various library functions.  See comments by
//  each function prototype to determine which of these may be returned.
#define SUCCESS                0
#define COMMAND_TIME_OUT       1
#define RESPONSE_TIME_OUT      2
#define INVALID_AXIS_SELECTION 3
#define MOVE_TIME_OUT          4
#define INVALID_PARAMETER      5
#define INVALID_BIT_NUMBER     6

// Controller status defintions
#define COMMAND_ERROR           0x01

/*Axis status definitions*/
#define AXIS_DIRECTION   0x01
#define AXIS_DONE        0x02
#define AXIS_OVERTRAVEL  0x04
#define AXIS_HOME_SWITCH 0x08

/*Encoder status definitions*/
#define SLIP_DETECT_ENABLED          0x01
#define POSITION_MAINTENANCE_ENABLED 0x02
#define AXIS_SLIPPED                 0x04
#define AXIS_WITHIN_DEADBAND         0x08
#define ENCODER_AT_HOME              0x10

/*Function arguments*/
#define POSITIVE    0
#define NEGATIVE    1
#define MODE_OFF    0
#define MODE_ON     1
#define OFF         0
#define ON          1
#define LOW         0
#define HIGH        1

/*Axis selection codes*/
#define OMS_X_AXIS   0x01
#define OMS_Y_AXIS   0x02
#define OMS_Z_AXIS   0x04
#define OMS_T_AXIS   0x08
#define OMS_U_AXIS   0x10
#define OMS_V_AXIS   0x20
#define OMS_R_AXIS   0x40
#define OMS_S_AXIS   0x80
#define OMS_ALL_AXES 0xff

/*I/O bit selection codes*/
#define BIT0  0x0001
#define BIT1  0x0002
#define BIT2  0x0004
#define BIT3  0x0008
#define BIT4  0x0010
#define BIT5  0x0020
#define BIT6  0x0040
#define BIT7  0x0080
#define BIT8  0x0100
#define BIT9  0x0200
#define BIT10 0x0400
#define BIT11 0x0800
#define BIT12 0x1000
#define BIT13 0x2000
#define BIT14 0x4000
#define BIT15 0x8000


#define MIN_VELOCITY 1
#define MAX_VELOCITY 4000000

#define MIN_ACCELERATION 1
#define MAX_ACCELERATION 8000000

typedef struct {
        long X;
        long Y;
        long Z;
        long T;
        long U;
        long V;
        long R;
        long S;
} AXES_DATA, *PAXES_DATA;

typedef struct {
        long Motor;
        long Encoder;
} AXIS_DATA;

typedef struct {
        AXIS_DATA X;
        AXIS_DATA Y;
        AXIS_DATA Z;
        AXIS_DATA T;
        AXIS_DATA U;
        AXIS_DATA V;
        AXIS_DATA R;
        AXIS_DATA S;
} POSITION_DATA, *PPOSITION_DATA;


//********************************************************************
//*             Alpabetic List of MAXp library functions             *
//*                                                                  *
//*    -> use GetOmsHandle() to open the controller                  *
//*    -> compile (link step) with "-lmaxp" switch to cause the      *
//*         linker (/bin/ld) to search /usr/lib/libmaxp for these    *
//*         functions                                                *
//********************************************************************

// OMS_GetHandle() --- open a specified device and return a handle to be
//    used on subsequent calls.  The string 'devname' should be of the
//    form "OmsMAXp1" through "OmsMAXp16" to specify a board number
//    from 1 to 16.  (Refer to documentation on setting the board number
//    via DIP switches.  Note that all boards installed must have a
//    unique board number.)
OMSHANDLE OMS_GetHandle(const char *devname);

// OMS_CloseHandle() --- close a device
void OMS_CloseHandle(OMSHANDLE OmsHandle);

// OMS_ResetController() --- Send a reset command to the
//    controller.  All parameters will be set to their default values and
//    the hardware is set to its power-up state.  Returns SUCCESS or
//    COMMAND_TIME_OUT.
long OMS_ResetController(OMSHANDLE OmsHandle);

// OMS_GetControllerDescription() --- copy the controller's firmware
//   description string in the buffer pointed to by argument
//   'Description'.  This string may be up to 128 bytes in length.
//   Returns SUCCESS or RESPONSE_TIME_OUT.
long OMS_GetControllerDescription(OMSHANDLE OmsHandle, char *Description);

// OMS_GetCmdBufferFree() --- return the number of bytes available in the
//    controller's command buffer.
long OMS_GetCmdBufferFree(OMSHANDLE OmsHandle);

// OMS_SendAndGetString() --- Send a string and get the response.
//  note that the 'response' pointer cannot be checked for sufficient
//   length to contain the response.  The library function assumes that
//   it can hold a response of 128 bytes.
long OMS_SendAndGetString(OMSHANDLE OmsHandle, const char *command,
                          char *response);

// OMS_SendString() --- send a string to the controller.  Returns SUCCESS or
//   or COMMAND_TIME_OUT.
long OMS_SendString(OMSHANDLE OmsHandle, const char *command);

// OMS_Wait() --- wait (Linux process sleep) for the specified number
//    of milliseconds, which must be a positive number (or zero) less than
//    or equal to 3600000.  In the event that the calling program handles
//    signals, they will continued to be handled during the sleep interval
//    (which will be resumed once the signal handler returns.)
//    Returns SUCCESS or INVALID_PARAMETER.
long OMS_Wait(long WaitTime);

// OMS_KillMotion() --- stop selected axes without using a deceleration
//    ramp and flush the command queue.
//    Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_KillMotion(OMSHANDLE OmsHandle, long AxisMask);

// OMS_StopMotion() --- stop selected axes using a deceleration ramp.
//    Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_StopMotion(OMSHANDLE OmsHandle, long AxisMask);

// OMS_GetMotorPosition() --- get selected axis motor positions, which are
//    stored into an instance of AXES_DATA referenced by argument 'lPosition'.
//    Returns SUCCESS or COMMAND_TIME_OUT.
long GetMotorPositions(OMSHANDLE OmsHandle,
                       long AxisMask,
                       PAXES_DATA lPosition);

// OMS_GetEncoderPosition() --- get selected axis encoder positions, which are
//    stored into an instance of AXES_DATA referenced by argument 'lPosition'.
//    Returns SUCCESS or COMMAND_TIME_OUT.
long GetEncoderPositions(OMSHANDLE OmsHandle,
                         long AxisMask,
                         PAXES_DATA lPosition);

// OMS_GetPositionData() --- get the motor and encoder positions for all axes
//    from the controller.  Positions are stored in structure referenced
//    by pointer argument 'lPosition'.  Function returns SUCCESS or
//    COMMAND_TIME_OUT.
long OMS_GetPositionData(OMSHANDLE OmsHandle,
		     PPOSITION_DATA lPosition);

// OMS_GetControllerFlags() --- get the controller error status flags
//    (written to the variable pointed to by reference argument 'pFlags'.
//    Returnss SUCCESS or COMMAND_TIME_OUT.
long OMS_GetControllerFlags(OMSHANDLE OmsHandle, long *pflags);

// OMS_ClrControllerFlags() --- clear specific controller error flags
//   Returns SUCCESS or COMMAND_TIME_OUT
long ClrOmsControllerFlags(OMSHANDLE OmsHandle, long flags);

// OMS_GetDoneFlags() --- return done status flags of all axes into the
//   reference variable, which will be set to an OR of zero or more of 
//   OMS_X_AXIS, etc..  
//   Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_GetDoneFlags(OMSHANDLE OmsHandle, long *pflags);

// OMS_ClrDoneFlags() --- clear DONE flags selected by a mask consisting
//    of an OR of one or more of OMS_X_AXIS, OMS_Y_AXIS, etc.
//    Returns SUCCESS or INVALID_AXIS_SELECTION.
long OMS_ClrDoneFlags(OMSHANDLE OmsHandle, long flags);

// OMS_GetPosLimitFlags() --- return positive limit status flags of all 
//   axes into the reference variable, which will be set to an OR of zero
//   or more of OMS_X_AXIS, etc..  
//   Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_GetPosLimitFlags(OMSHANDLE OmsHandle, long *pflags);

// OMS_ClrPosLimitFlags() --- clear positive limit flags selected by a
//    mask consisting of an OR of one or more of OMS_X_AXIS, OMS_Y_AXIS, etc.
//    Returns SUCCESS or INVALID_AXIS_SELECTION.
long OMS_ClrPosLimitFlags(OMSHANDLE OmsHandle, long flags);

// OMS_GetNegLimitFlags() --- return negative limit status flags of all 
//   axes into the reference variable, which will be set to an OR of zero
//   or more of OMS_X_AXIS, etc..  
//   Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_GetNegLimitFlags(OMSHANDLE OmsHandle, long *pflags);

// OMS_ClrNegLimitFlags() --- clear negative limit flags selected by a
//    mask consisting of an OR of one or more of OMS_X_AXIS, OMS_Y_AXIS, etc.
//    Returns SUCCESS or INVALID_AXIS_SELECTION.
long OMS_ClrNegLimitFlags(OMSHANDLE OmsHandle, long flags);

// OMS_GetSlipFlags() --- return slip status flags of all axes into the
//   reference variable, which will be set to an OR of zero or more of 
//   OMS_X_AXIS, etc..  
//   Returns SUCCESS or COMMAND_TIME_OUT.
long OMS_GetSlipFlags(OMSHANDLE OmsHandle, long *pflags);

// OMS_ClrSlipFlags() --- clear slip flags selected by a mask consisting
//    of an OR of one or more of OMS_X_AXIS, OMS_Y_AXIS, etc.
//    Returns SUCCESS or INVALID_AXIS_SELECTION.
long OMS_ClrSlipFlags(OMSHANDLE OmsHandle, long flags);

// OMS_GetAxisFlags() -- obtain status information about a single axis.
//   The flag status information is encoded:
//      bit 0 -- DIRECTION    0=Positive 1=Negative
//      bit 1 -- DONE         0=False    1=True
//      bit 2 -- OVERTRAVEL   0=False    1=True
//      bit 3 -- HOME_SWITCH  0=False    1=True
//
//   This information is returned via a reference pointer given in
//   argument FlagStatus.  The function returns SUCCESS,
//   INVALID_AXIS_SELECTION, or RESPONSE_TIME_OUT.
long OMS_GetAxisFlags(OMSHANDLE OmsHandle,
		      long   AxisSelection,  // ONE of OMS_X_AXIS, etc.
		      long  *FlagStatus);

// OMS_GetEncoderFlags() --- return the encoder status flags of a
//   selected axis.  The flags are returned via a reference pointer
//   (argument FlagStatus), and are encoded:
//      Bit 0 - SLIP DETECT ENABLE           0=Disabled         1=Enabled
//      Bit 1 - POSITION MAINTENANCE ENABLE  0=Disabled         1=Enabled
//      Bit 2 - SLIP OR STALL DETECTED       0=Not detected     1=Detected
//      Bit 3 - WITHIN DEADBAND              0=Not in deadband  1=In deadband
//      Bit 4 - AXIS HOME                    0=Not at home      1=At home
//   Function returns SUCCESS, INVALID_AXIS_SELECTION, RESPONSE_TIME_OUT.
long OMS_GetEncoderFlags(OMSHANDLE  OmsHandle,
			long    AxisSelection,
			long   *FlagStatus);

#endif // __OMS_MAXP_LIB_H_
